"""See `PathTemplate`."""

from __future__ import annotations

import enum
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, assert_never

import attrs

from usdb_syncer import errors, utils
from usdb_syncer.custom_data import CustomData

if TYPE_CHECKING:
    from usdb_syncer.usdb_song import UsdbSong

FORBIDDEN_CHARACTERS = '?"<>|*.'
UNKNOWN_PLACEHOLDER_STRING = "None"


class PathTemplateError(errors.UsdbSyncerError, ValueError):
    """Raised when the path template is invalid."""


@attrs.define
class InvalidCharError(PathTemplateError):
    """Raised when the path template contains a forbidden character."""

    char: str

    def __str__(self) -> str:
        return f"Invalid character in path template: '{self.char}'"


@attrs.define
class InvalidPlaceholderError(PathTemplateError):
    """Raised when the path template contains an unknown placeholder name."""

    name: str

    def __str__(self) -> str:
        return f"Invalid placeholder in path template: '{self.name}'"


@attrs.define
class NotEnoughComponentsError(PathTemplateError):
    """Raised when the path template contains less than two components."""

    def __str__(self) -> str:
        return "Path template must contain at least two components separated by '/'!"


@attrs.define
class PathTemplate:
    """A path with optional placeholder names, which can be resolved by passing a
    UsdbSong object. The syntax for placeholders is `:name:`.
    See `PathTemplatePlaceholder` for valid names.
    """

    _components: list[PathTemplateComponent]
    default_str = ":artist: - :title: / :artist: - :title:"

    @classmethod
    def parse(cls, template: str) -> PathTemplate:
        parts = [
            p for part in template.replace("\\", "/").split("/") if (p := part.strip())
        ]
        if len(parts) < 2:
            raise NotEnoughComponentsError
        return cls([PathTemplateComponent.parse(part) for part in parts])

    def evaluate(self, song: UsdbSong, parent: Path = Path()) -> Path:
        """Returns a valid path relative to `parent` with placeholders replaced with
        the values from `song`. The final component is the filename stem.
        """
        return Path(
            parent,
            *(utils.sanitize_filename(c.evaluate(song)) for c in self._components),
        )

    @classmethod
    def default(cls) -> PathTemplate:
        return cls.parse(cls.default_str)

    def __str__(self) -> str:
        return " / ".join(map(str, self._components))


@attrs.define
class PathTemplateComponent:
    """A component of a template path, i.e. a file or directory name supporting the
    template syntax.
    """

    _tokens: list[PathTemplateComponentToken]

    @classmethod
    def parse(cls, component: str) -> PathTemplateComponent:
        component = component.strip()
        if component.count(":") % 2 == 1:
            raise InvalidCharError(":")
        tokens: list[PathTemplateComponentToken] = []
        literal = True
        for token in component.split(":"):
            if literal:
                if token:
                    tokens.append(PathTemplateLiteral(token))
            else:
                if token[:1] == "*":
                    tokens.append(PathTemplateCustomPlaceholder(token[1:]))
                else:
                    tokens.append(PathTemplatePlaceholder.from_name(token))
            literal = not literal
        return cls(tokens)

    def evaluate(self, song: UsdbSong) -> str:
        return "".join(t.evaluate(song) for t in self._tokens)

    def __str__(self) -> str:
        return "".join(map(str, self._tokens))


class PathTemplateComponentToken:
    """Common base class for path template component tokens."""

    def evaluate(self, _song: UsdbSong) -> str:
        return NotImplemented


class PathTemplateLiteral(PathTemplateComponentToken, str):
    """A literal string that is part of a path template."""

    def __init__(self, literal: str) -> None:
        for char in FORBIDDEN_CHARACTERS:
            if char in literal:
                raise InvalidCharError(char)

    def evaluate(self, _song: UsdbSong) -> str:
        return str(self)


class PathTemplatePlaceholder(PathTemplateComponentToken, enum.Enum):
    """The supported placeholders for path templates."""

    SONG_ID = "id"
    ARTIST = "artist"
    ARTIST_INITIAL = "artist initial"
    TITLE = "title"
    GENRE = "genre"
    YEAR = "year"
    LANGUAGE = "language"
    CREATOR = "creator"
    EDITION = "edition"
    RATING = "rating"

    @classmethod
    def from_name(cls, name: str) -> PathTemplatePlaceholder:
        try:
            return PathTemplatePlaceholder(name)
        except ValueError as error:
            raise InvalidPlaceholderError(name) from error

    def evaluate(self, song: UsdbSong) -> str:  # noqa: C901
        match self:
            case PathTemplatePlaceholder.SONG_ID:
                return str(song.song_id)
            case PathTemplatePlaceholder.ARTIST:
                if song.artist and len(song.artist) > 0:
                    return song.artist
                return UNKNOWN_PLACEHOLDER_STRING
            case PathTemplatePlaceholder.ARTIST_INITIAL:
                return (
                    utils.get_first_alphanum_upper(song.artist)
                    or UNKNOWN_PLACEHOLDER_STRING
                )
            case PathTemplatePlaceholder.TITLE:
                if song.title and len(song.title) > 0:
                    return song.title
                return UNKNOWN_PLACEHOLDER_STRING
            case PathTemplatePlaceholder.YEAR:
                if song.year and song.year > 0:
                    return str(song.year)
                return UNKNOWN_PLACEHOLDER_STRING
            case PathTemplatePlaceholder.GENRE:
                return next(iter(song.genres()), UNKNOWN_PLACEHOLDER_STRING)
            case PathTemplatePlaceholder.LANGUAGE:
                return next(iter(song.languages()), UNKNOWN_PLACEHOLDER_STRING)
            case PathTemplatePlaceholder.CREATOR:
                return next(iter(song.creators()), UNKNOWN_PLACEHOLDER_STRING)
            case PathTemplatePlaceholder.EDITION:
                if song.edition and len(song.edition) > 0:
                    return song.edition
                return UNKNOWN_PLACEHOLDER_STRING
            case PathTemplatePlaceholder.RATING:
                # This is annoying because we can't differentiate between a rating of 0
                # and no rating. Would take a refactor of UsdbSong to fix.
                if song.rating:
                    return str(song.rating)
                return UNKNOWN_PLACEHOLDER_STRING
            case _ as unreachable:
                assert_never(unreachable)

    def __str__(self) -> str:
        return f":{self.value}:"


class PathTemplateCustomPlaceholder(PathTemplateComponentToken):
    """A path template placeholder representing a custom data key."""

    _key: str

    def __init__(self, key: str) -> None:
        if not CustomData.is_valid_key(key):
            raise InvalidPlaceholderError(key)
        self._key = key

    def evaluate(self, song: UsdbSong) -> str:
        if song.sync_meta:
            return song.sync_meta.custom_data.get(self._key) or ""
        return ""

    def __str__(self) -> str:
        return f":*{self._key}:"

    @classmethod
    def options(cls) -> Iterable[PathTemplateCustomPlaceholder]:
        return (cls(k) for k in CustomData.key_options())



"""Model for the filter tree."""

from __future__ import annotations

from typing import Any, cast, overload

from PySide6.QtCore import (
    QAbstractItemModel,
    QModelIndex,
    QObject,
    QPersistentModelIndex,
    QSortFilterProxyModel,
    Qt,
    QTimer,
)
from PySide6.QtWidgets import QWidget

from usdb_syncer import db

from .item import Filter, FilterItem, RootItem, SavedSearch, TreeItem, VariantItem

QIndex = QModelIndex | QPersistentModelIndex


class TreeModel(QAbstractItemModel):
    """Model for the filter tree."""

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.root = RootItem()
        self.root.set_children(FilterItem(data=f, parent=self.root) for f in Filter)
        # saved searches is not checkable
        self.root.children[0].checked = None

    def item_for_index(self, idx: QIndex) -> TreeItem | None:
        if idx.isValid():
            return cast(TreeItem, idx.internalPointer())
        return None

    def index_for_item(self, item: TreeItem) -> QModelIndex:
        return self.createIndex(item.row_in_parent, 0, item)

    def emit_item_changed(self, item: TreeItem) -> None:
        idx = self.index_for_item(item)
        self.dataChanged.emit(idx, idx, [Qt.ItemDataRole.CheckStateRole])

    # change data

    def populate(self) -> None:
        self.beginResetModel()
        for item in self.root.children:
            item.set_children(
                VariantItem(data=var, parent=item, checkable=item.checkable)
                for var in item.data.variants()
            )
        self.endResetModel()

    def set_checked(self, item: TreeItem, checked: bool) -> None:
        if checked and item.parent:
            for sibling in item.parent.children:
                sibling.checked = False

        item.checked = checked

    def insert_saved_search(self, search: SavedSearch) -> QModelIndex:
        with db.transaction():
            search.insert()
        parent = self.root.children[0]
        parent_idx = self.index_for_item(parent)
        self.beginInsertRows(parent_idx, 0, 0)
        self.root.children[0].children = (
            VariantItem(data=search, parent=parent),
            *self.root.children[0].children,
        )
        self.endInsertRows()
        return self.index(0, 0, parent_idx)

    def delete_saved_search(self, index: QModelIndex) -> None:
        item = self.item_for_index(index)
        if not item or not isinstance(item.data, SavedSearch) or not item.parent:
            return
        self.beginRemoveRows(index.parent(), index.row(), index.row())
        with db.transaction():
            item.data.delete()
        children = item.parent.children
        item.parent.children = (*children[: index.row()], *children[index.row() + 1 :])
        self.endRemoveRows()

    # QAbstractItemModel implementation

    def rowCount(self, parent: QIndex | None = None) -> int:  # noqa: N802
        if parent is None:
            parent = QModelIndex()
        if not parent.isValid():
            return len(self.root.children)
        item = cast(TreeItem, parent.internalPointer())
        return len(item.children)

    def columnCount(self, parent: QIndex | None = None) -> int:  # noqa: N802
        return 1

    def index(self, row: int, column: int, parent: QIndex | None = None) -> QModelIndex:
        if parent is None:
            parent = QModelIndex()
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        if parent.isValid():
            parent_item = cast(TreeItem, parent.internalPointer())
        else:
            parent_item = self.root
        item = parent_item.children[row]
        return self.createIndex(row, column, item)

    @overload  # type: ignore
    def parent(self, child: QIndex) -> QModelIndex: ...

    @overload
    def parent(self) -> QObject: ...

    def parent(self, child: QIndex | None = None) -> QModelIndex | QObject:
        if child is None:
            return super().parent()
        if not child.isValid():
            return QModelIndex()
        child_item = cast(TreeItem, child.internalPointer())
        parent_item = child_item.parent
        if parent_item is None or parent_item is self.root:
            return QModelIndex()
        return self.createIndex(parent_item.row_in_parent, 0, parent_item)

    def data(self, index: QIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if not (item := self.item_for_index(index)):
            return None

        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            return str(item.data)
        if role == Qt.ItemDataRole.CheckStateRole:
            return item.checked if item.checkable else None
        if role == Qt.ItemDataRole.DecorationRole:
            return item.decoration()
        return None

    def flags(self, index: QIndex) -> Qt.ItemFlag:
        if item := self.item_for_index(index):
            return item.flags()
        return Qt.ItemFlag.NoItemFlags

    def setData(self, index: QIndex, value: Any, role: int = 0) -> bool:  # noqa: N802
        if not (
            role == Qt.ItemDataRole.EditRole
            and value
            and isinstance(value, str)
            and (item := self.item_for_index(index))
            and isinstance(item.data, SavedSearch)
            and item.data.name != value
        ):
            return False
        item.data.update(new_name=value)
        return True


class TreeProxyModel(QSortFilterProxyModel):
    """Proxy model for filtering the filter tree by text."""

    def __init__(self, parent: QObject, source_model: TreeModel) -> None:
        super().__init__(parent)
        self._source = source_model
        self.setSourceModel(source_model)
        self._filter: str = ""
        self._matches: dict[Filter, set[str | int]] = {}
        self._filter_invalidation_timer = QTimer(parent)
        self._filter_invalidation_timer.setSingleShot(True)
        self._filter_invalidation_timer.setInterval(400)
        self._filter_invalidation_timer.timeout.connect(self._on_filter_changed)

    def filterAcceptsRow(  # noqa: N802
        self, source_row: int, source_parent: QIndex
    ) -> bool:
        if not self._filter or not (
            parent := self._source.item_for_index(source_parent)
        ):
            return True
        return parent.children[source_row].is_accepted(self._matches)

    def set_filter(self, text: str) -> None:
        if (new := text.strip()) != self._filter:
            self._filter = new
            self._filter_invalidation_timer.start()

    def _on_filter_changed(self) -> None:
        if self._filter:
            self._matches = {
                Filter.ARTIST: set(db.search_usdb_song_artists(self._filter)),
                Filter.TITLE: set(db.search_usdb_song_titles(self._filter)),
                Filter.EDITION: set(db.search_usdb_song_editions(self._filter)),
                Filter.LANGUAGE: set(db.search_usdb_song_languages(self._filter)),
                Filter.YEAR: set(db.search_usdb_song_years(self._filter)),
                Filter.GENRE: set(db.search_usdb_song_genres(self._filter)),
                Filter.CREATOR: set(db.search_usdb_song_creators(self._filter)),
            }
        else:
            self._matches = {}
        self.invalidateRowsFilter()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'UsdbLoginDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QComboBox, QCommandLinkButton,
    QDialog, QDialogButtonBox, QFormLayout, QHBoxLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(294, 214)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.label = QLabel(Dialog)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.label)

        self.line_edit_username = QLineEdit(Dialog)
        self.line_edit_username.setObjectName(u"line_edit_username")

        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.line_edit_username)

        self.label_2 = QLabel(Dialog)
        self.label_2.setObjectName(u"label_2")

        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.label_2)

        self.line_edit_password = QLineEdit(Dialog)
        self.line_edit_password.setObjectName(u"line_edit_password")
        self.line_edit_password.setEchoMode(QLineEdit.Password)

        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.line_edit_password)

        self.label_5 = QLabel(Dialog)
        self.label_5.setObjectName(u"label_5")

        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.label_5)

        self.combobox_browser = QComboBox(Dialog)
        self.combobox_browser.setObjectName(u"combobox_browser")

        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.combobox_browser)


        self.verticalLayout.addLayout(self.formLayout)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.command_link_register = QCommandLinkButton(Dialog)
        self.command_link_register.setObjectName(u"command_link_register")

        self.horizontalLayout_2.addWidget(self.command_link_register)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.button_check_login = QPushButton(Dialog)
        self.button_check_login.setObjectName(u"button_check_login")

        self.horizontalLayout.addWidget(self.button_check_login)

        self.button_log_out = QPushButton(Dialog)
        self.button_log_out.setObjectName(u"button_log_out")

        self.horizontalLayout.addWidget(self.button_log_out)

        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Save)

        self.horizontalLayout.addWidget(self.buttonBox)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"USDB Login", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Username:", None))
#if QT_CONFIG(tooltip)
        self.line_edit_username.setToolTip(QCoreApplication.translate("Dialog", u"Enter your USDB credentials to automatically log in if there is no existing session.", None))
#endif // QT_CONFIG(tooltip)
        self.label_2.setText(QCoreApplication.translate("Dialog", u"Password:", None))
#if QT_CONFIG(tooltip)
        self.line_edit_password.setToolTip(QCoreApplication.translate("Dialog", u"Enter your USDB credentials to automatically log in if there is no existing session.", None))
#endif // QT_CONFIG(tooltip)
        self.label_5.setText(QCoreApplication.translate("Dialog", u"Browser cookies:", None))
#if QT_CONFIG(tooltip)
        self.combobox_browser.setToolTip(QCoreApplication.translate("Dialog", u"Select your browser in order to reuse an existing USDB session.", None))
#endif // QT_CONFIG(tooltip)
        self.command_link_register.setText(QCoreApplication.translate("Dialog", u"No account?", None))
        self.command_link_register.setDescription(QCoreApplication.translate("Dialog", u"Register on USDB", None))
        self.button_check_login.setText(QCoreApplication.translate("Dialog", u"Check Login", None))
        self.button_log_out.setText(QCoreApplication.translate("Dialog", u"Log Out", None))
    # retranslateUi




"""General-purpose utilities for the GUI."""

from collections.abc import Callable
from typing import Any

import attrs
from PySide6.QtCore import QObject, Qt
from PySide6.QtGui import QKeySequence, QShortcut
from PySide6.QtWidgets import QAbstractScrollArea, QApplication


@attrs.define(kw_only=True)
class Modifiers:
    """Pressed keyboard modifiers."""

    ctrl: bool
    shift: bool
    alt: bool


def set_shortcut(key: str, parent: QObject, func: Callable[[], Any]) -> None:
    QShortcut(QKeySequence(key), parent).activated.connect(func)


def scroll_to_bottom(scroll_area: QAbstractScrollArea) -> None:
    slider = scroll_area.verticalScrollBar()
    slider.setValue(slider.maximum())


def keyboard_modifiers() -> Modifiers:
    mods = QApplication.keyboardModifiers()
    return Modifiers(
        ctrl=bool(mods & Qt.KeyboardModifier.ControlModifier),
        shift=bool(mods & Qt.KeyboardModifier.ShiftModifier),
        alt=bool(mods & Qt.KeyboardModifier.AltModifier),
    )



"""Dialog for adding custom sync meta data."""

from collections.abc import Callable

from PySide6 import QtWidgets

from usdb_syncer.custom_data import CustomData
from usdb_syncer.gui.forms.CustomDataDialog import Ui_Dialog


class CustomDataDialog(Ui_Dialog, QtWidgets.QDialog):
    """Dialog with about info and credits."""

    def __init__(
        self,
        parent: QtWidgets.QWidget,
        on_accept: Callable[[str, str], None],
        key: str | None = None,
    ) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self._on_accept = on_accept
        if key:
            self.edit_key.setText(key)

    def accept(self) -> None:
        key = self.edit_key.text().strip()
        value = self.edit_value.text().strip()
        if not key or not value:
            warning = "Both key and value must be supplied!"
            QtWidgets.QMessageBox.warning(
                self, "Warning", "Both key and value must be supplied!"
            )
        elif not CustomData.is_valid_key(key):
            warning = (
                "Key must not contain any of these characters: "
                + CustomData.FORBIDDEN_CHARACTERS
            )
        else:
            self._on_accept(key, value)
            super().accept()
            return
        QtWidgets.QMessageBox.warning(self, "Warning", warning)



"""Dictionary of language names and their English translations"""

LANGUAGE_TRANSLATIONS = {
    # Arabic
    "arabic": "Arabic",
    "arabisch": "Arabic",
    "arabe": "Arabic",
    "árabe": "Arabic",
    "arabo": "Arabic",
    "arabski": "Arabic",
    # Chinese
    "chinese": "Chinese",
    "chinesisch": "Chinese",
    "chinois": "Chinese",
    "chino": "Chinese",
    "cinese": "Chinese",
    "chinês": "Chinese",
    "chiński": "Chinese",
    # Croatian
    "croatian": "Croatian",
    "kroatisch": "Croatian",
    "croate": "Croatian",
    "croata": "Croatian",
    "croato": "Croatian",
    "chorwacki": "Croatian",
    # Czech
    "czech": "Czech",
    "tschechisch": "Czech",
    "tchèque": "Czech",
    "ceco": "Czech",
    "checo": "Czech",
    "czeski": "Czech",
    # Danish
    "danish": "Danish",
    "dänisch": "Danish",
    "danois": "Danish",
    "danés": "Danish",
    "danese": "Danish",
    "dinamarquês": "Danish",
    "duński": "Danish",
    # Dutch
    "dutch": "Dutch",
    "holländisch": "Dutch",
    "hollandais": "Dutch",
    "holandés": "Dutch",
    "olandese": "Dutch",
    "holandês": "Dutch",
    "holenderski": "Dutch",
    # English
    "english": "English",
    "englisch": "English",
    "anglais": "English",
    "inglés": "English",
    "inglese": "English",
    "inglês": "English",
    "angielski": "English",
    # Finnish
    "finnish": "Finnish",
    "finnisch": "Finnish",
    "finois": "Finnish",
    "finlandés": "Finnish",
    "finlandese": "Finnish",
    "finlandês": "Finnish",
    "fiński": "Finnish",
    # French
    "french": "French",
    "französisch": "French",
    "français": "French",
    "francés": "French",
    "francese": "French",
    "francês": "French",
    "francuski": "French",
    # German
    "german": "German",
    "deutsch": "German",
    "allemand": "German",
    "alemán": "German",
    "tedesco": "German",
    "alemão": "German",
    "niemiecki": "German",
    # Hindi
    "hindi": "Hindi",
    "indisch": "Hindi",
    "indien": "Hindi",
    "indio": "Hindi",
    "indiano": "Hindi",
    "indyjski": "Hindi",
    # Italian
    "italian": "Italian",
    "italienisch": "Italian",
    "italien": "Italian",
    "italiano": "Italian",
    "włoski": "Italian",
    # Japanese
    "japanese": "Japanese",
    "japanisch": "Japanese",
    "japonais": "Japanese",
    "japonés": "Japanese",
    "giapponese": "Japanese",
    "japonês": "Japanese",
    "japoński": "Japanese",
    # Korean
    "korean": "Korean",
    "koreanisch": "Korean",
    "coréen": "Korean",
    "coreano": "Korean",
    "koreański": "Korean",
    # Latin
    "lateinisch": "Latin",
    "latin": "Latin",
    "latino": "Latin",
    "łacina": "Latin",
    # Norwegian
    "norwegian": "Norwegian",
    "norwegisch": "Norwegian",
    "norvégien": "Norwegian",
    "noruego": "Norwegian",
    "norvegese": "Norwegian",
    "norueguês": "Norwegian",
    "norweski": "Norwegian",
    # Polish
    "polish": "Polish",
    "polnisch": "Polish",
    "polonais": "Polish",
    "polacco,": "Polish",
    "polaco": "Polish",
    "polski": "Polish",
    # Portuguese
    "portuguese": "Portuguese",
    "portugiesisch": "Portuguese",
    "portugais": "Portuguese",
    "portugués": "Portuguese",
    "portoghese": "Portuguese",
    "português": "Portuguese",
    "portugalski": "Portuguese",
    # Romanian
    "romanian": "Romanian",
    "rumänisch": "Romanian",
    "roumain": "Romanian",
    "rumano": "Romanian",
    "rumeno": "Romanian",
    "romeno": "Romanian",
    "rumuński": "Romanian",
    # Russian
    "russian": "Russian",
    "russisch": "Russian",
    "russe": "Russian",
    "ruso": "Russian",
    "russo": "Russian",
    "rosyjski": "Russian",
    "ruski": "Russian",
    # Slovak
    "slovak": "Slovak",
    "slowakisch": "Slovak",
    "slovaque": "Slovak",
    "eslovaco": "Slovak",
    "slovacco": "Slovak",
    "słowacki": "Slovak",
    # Slowenian
    "slowenian": "Slowenian",
    "slowenisch": "Slowenian",
    "slovène": "Slowenian",
    "esloveno": "Slowenian",
    "sloveno": "Slowenian",
    "słoweński": "Slowenian",
    # Spanish
    "spanish": "Spanish",
    "spanisch": "Spanish",
    "espagnol": "Spanish",
    "español": "Spanish",
    "spagnolo": "Spanish",
    "espanhol": "Spanish",
    "hiszpański": "Spanish",
    # Swedish
    "swedish": "Swedish",
    "schwedisch": "Swedish",
    "suédois": "Swedish",
    "sueco": "Swedish",
    "svedese": "Swedish",
    "szwedzki": "Swedish",
    # Turkish
    "turkish": "Turkish",
    "türkisch": "Turkish",
    "turc": "Turkish",
    "turco": "Turkish",
    "turecki ": "Turkish",
}



"""Constants"""

import re

MINIMUM_BPM = 200.0
GITHUB_API_LATEST = "https://api.github.com/repos/bohning/usdb_syncer/releases/latest"
GITHUB_DL_LATEST = "https://github.com/bohning/usdb_syncer/releases/latest"
GITHUB_SHARED_CONTENT = (
    "https://raw.githubusercontent.com/bohning/usdb_syncer/refs/heads/main/shared"
)
DISCORD_WEBHOOK_API = "https://discordapp.com/api/webhooks"


class UsdbStrings:
    """Relevant strings from USDB"""

    WELCOME: str
    SONG_LANGUAGE: str
    SONG_YEAR: str
    SONG_EDITED_BY: str
    SONG_RATING: str
    GOLDEN_NOTES: str
    SONGCHECK: str
    DATE: str
    UPLOADED_BY: str
    VIEWS: str
    YES: str
    NO: str
    DATASET_NOT_FOUND = "Datensatz nicht gefunden"
    WELCOME_PLEASE_LOGIN = "Welcome, Please login"
    LOGIN_INVALID = "Login or Password invalid, please try again."
    NOT_LOGGED_IN = "You are not logged in. Login to use this function."


class UsdbStringsEnglish(UsdbStrings):
    """Relevant strings from USDB"""

    WELCOME = "Welcome"
    SONG_LANGUAGE = "Language"
    SONG_YEAR = "Year"
    SONG_EDITED_BY = "Song edited by:"
    SONG_RATING = "Rating"
    GOLDEN_NOTES = "Golden Notes"
    SONGCHECK = "Songcheck"
    DATE = "Date"
    UPLOADED_BY = "Uploaded by"
    VIEWS = "Views"
    YES = "Yes"
    NO = "No"


class UsdbStringsGerman(UsdbStrings):
    """Relevant strings from USDB"""

    WELCOME = "Willkommen"
    SONG_LANGUAGE = "Sprache"
    SONG_YEAR = "Jahr"
    SONG_EDITED_BY = "Song editiert von:"
    SONG_RATING = "Bewertung"
    GOLDEN_NOTES = "Goldene Noten"
    SONGCHECK = "Songcheck"
    DATE = "Datum"
    UPLOADED_BY = "Hochgeladen von"
    VIEWS = "Aufrufe"
    YES = "Ja"
    NO = "Nein"


class UsdbStringsFrench(UsdbStrings):
    """Relevant strings from USDB"""

    WELCOME = "Bienvenue"
    SONG_LANGUAGE = "Langue"
    SONG_YEAR = "An"
    SONG_EDITED_BY = "Chanson modifiée par:"
    SONG_RATING = "Classement"
    GOLDEN_NOTES = "Notes en or"
    SONGCHECK = "Songcheck"
    DATE = "Date"
    UPLOADED_BY = "Téléchargé par"
    VIEWS = "Affichages"
    YES = "Oui"
    NO = "Non"


class Usdb:
    """Constants related to USDB."""

    DOMAIN = "usdb.animux.de"
    BASE_URL = "https://" + DOMAIN + "/"
    REGISTER_URL = BASE_URL + "index.php?link=register"
    GETTXT_URL = BASE_URL + "?link=gettxt&id="
    DETAIL_URL = BASE_URL + "?link=detail&id="
    COVER_URL = BASE_URL + "data/cover/"
    MAX_SONG_ID = 100_000
    MAX_SONGS_PER_PAGE = 100
    DATETIME_STRF = "%d.%m.%y - %H:%M"


class YtErrorMsg:
    """Strings returned by yt-dlp when download fails"""

    YT_AGE_RESTRICTED = (
        "Sign in to confirm your age. This video may be inappropriate for some users."
    )
    YT_GEO_RESTRICTED = (
        "Video unavailable. The uploader has not made this video available in your "
        "country"
    )
    YT_UNAVAILABLE = "Video unavailable"
    YT_PARSE_ERROR = "Failed to parse XML"
    YT_FORBIDDEN = "HTTP Error 403: Forbidden"


SUPPORTED_VIDEO_SOURCES_REGEX = re.compile(
    r"""\b
        (
            (?:https?://)?
            (?:www\.)?
            (?:
                youtube\.com
                | youtube-nocookie\.com
                | youtu\.be
                | vimeo\.com
                | web\.archive\.org
                | fb\.watch
                | universal-music\.de
                | dailymotion\.com
            )
            /\S+
        )
    """,
    re.VERBOSE,
)

ISO_639_2B_LANGUAGE_CODES = {
    "Bosnian": "bos",
    "Breton": "bre",
    "Catalan": "cat",
    "Chinese": "chi",
    "Croatian": "hrv",
    "Czech": "cze",
    "Danish": "dan",
    "Duala": "dua",
    "Dutch": "dut",
    "English": "eng",
    "Estonian": "est",
    "Finnish": "fin",
    "French": "fre",
    "Galician": "glg",
    "German": "ger",
    "Haitian": "hat",
    "Hebrew": "heb",
    "Hindi": "hin",
    "Hungarian": "hun",
    "Icelandic": "ice",
    "Indonesian": "ind",
    "Irish": "gle",
    "Italian": "ita",
    "Japanese": "jpn",
    "Korean": "kor",
    "Latin": "lat",
    "Malagasy": "mlg",
    "Maori": "mao",
    "Norwegian": "nor",
    "Polish": "pol",
    "Portuguese": "por",
    "Quechua": "que",
    "Romanian": "rum",
    "Russian": "rus",
    "Samoan": "smo",
    "Scots": "sco",
    "Serbian": "srp",
    "Slovak": "slo",
    "Slovenian": "slv",
    "Spanish": "spa",
    "Swedish": "swe",
    "Tagalog": "tgl",
    "Turkish": "tur",
    "Ukrainian": "ukr",
    "Vietnamese": "vie",
    "Welsh": "wel",
    "Yoruba": "yor",
    "Zulu": "zul",
}

QUOTATION_MARKS = {
    "Bosnian": ("”", "”"),
    "Breton": ("«", "»"),
    "Catalan": ("«", "»"),
    "Chinese (romanized)": ("“", "”"),
    "Croatian": ("„", "”"),
    "Czech": ("„", "“"),
    "Danish": ("»", "«"),
    "Dutch": ("„", "”"),
    "English": ("“", "”"),
    "Estonian": ("„", "“"),
    "Finnish": ("”", "”"),
    "French": ("«", "»"),
    "Galician": ("«", "»"),
    "German": ("»", "«"),
    "Greek": ("«", "»"),
    "Greek (romanized)": ("«", "»"),
    "Hebrew": ("„", "”"),
    "Hindi": ("“", "”"),
    "Hindi (romanized)": ("“", "”"),
    "Hungarian": ("„", "”"),
    "Icelandic": ("„", "“"),
    "Indonesian": ("“", "”"),
    "Irish": ("“", "”"),
    "Italian": ("«", "»"),
    "Japanese": ("「", "」"),
    "Japanese (romanized)": ("「", "」"),
    "Korean": ("“", "”"),
    "Korean (romanized)": ("“", "”"),
    "Norwegian": ("«", "»"),
    "Polish": ("„", "”"),
    "Portuguese": ("«", "»"),
    "Portuguese (Brazil)": ("“", "”"),
    "Romanian": ("„", "”"),
    "Russian": ("«", "»"),
    "Russian (romanized)": ("«", "»"),
    "Scots": ("“", "”"),
    "Serbian": ("„", "“"),
    "Slovak": ("„", "“"),
    "Slovenian": ("„", "“"),
    "Spanish": ("«", "»"),
    "Swedish": ("”", "”"),
    "Turkish": ("“", "”"),
    "Ukrainian": ("«", "»"),
    "Vietnamese": ("“", "”"),
    "Welsh": ("‘", "’"),  # noqa: RUF001
}

QUOTATION_MARKS_TO_REPLACE = {'"', "„", "”", "«", "»", "“", "「", "」"}

LANGUAGES_WITH_SPACED_QUOTES = {"French", "Breton"}



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'SettingsDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCheckBox, QComboBox,
    QDialog, QDialogButtonBox, QGridLayout, QGroupBox,
    QHBoxLayout, QLabel, QLineEdit, QPushButton,
    QSizePolicy, QSpacerItem, QSpinBox, QTabWidget,
    QVBoxLayout, QWidget)
from usdb_syncer.gui.resources.qt import resources as resources_rc

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(673, 828)
        self.gridLayout = QGridLayout(Dialog)
        self.gridLayout.setObjectName(u"gridLayout")
        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Apply|QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)

        self.gridLayout.addWidget(self.buttonBox, 3, 0, 1, 2)

        self.tabWidget = QTabWidget(Dialog)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.verticalLayout_6 = QVBoxLayout(self.tab)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.horizontalLayout_11 = QHBoxLayout()
        self.horizontalLayout_11.setObjectName(u"horizontalLayout_11")
        self.label_theme = QLabel(self.tab)
        self.label_theme.setObjectName(u"label_theme")

        self.horizontalLayout_11.addWidget(self.label_theme)

        self.comboBox_theme = QComboBox(self.tab)
        self.comboBox_theme.setObjectName(u"comboBox_theme")

        self.horizontalLayout_11.addWidget(self.comboBox_theme)


        self.verticalLayout_6.addLayout(self.horizontalLayout_11)

        self.horizontalLayout_14 = QHBoxLayout()
        self.horizontalLayout_14.setObjectName(u"horizontalLayout_14")
        self.label_primary_color = QLabel(self.tab)
        self.label_primary_color.setObjectName(u"label_primary_color")

        self.horizontalLayout_14.addWidget(self.label_primary_color)

        self.comboBox_primary_color = QComboBox(self.tab)
        self.comboBox_primary_color.setObjectName(u"comboBox_primary_color")

        self.horizontalLayout_14.addWidget(self.comboBox_primary_color)


        self.verticalLayout_6.addLayout(self.horizontalLayout_14)

        self.horizontalLayout_12 = QHBoxLayout()
        self.horizontalLayout_12.setObjectName(u"horizontalLayout_12")
        self.label_3 = QLabel(self.tab)
        self.label_3.setObjectName(u"label_3")

        self.horizontalLayout_12.addWidget(self.label_3)

        self.checkBox_colored_background = QCheckBox(self.tab)
        self.checkBox_colored_background.setObjectName(u"checkBox_colored_background")

        self.horizontalLayout_12.addWidget(self.checkBox_colored_background)


        self.verticalLayout_6.addLayout(self.horizontalLayout_12)

        self.verticalSpacer_4 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_6.addItem(self.verticalSpacer_4)

        self.tabWidget.addTab(self.tab, "")
        self.tabDownload = QWidget()
        self.tabDownload.setObjectName(u"tabDownload")
        self.horizontalLayout = QHBoxLayout(self.tabDownload)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(20, -1, -1, -1)
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.groupBox_login = QGroupBox(self.tabDownload)
        self.groupBox_login.setObjectName(u"groupBox_login")
        self.gridLayout_3 = QGridLayout(self.groupBox_login)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.label_login_cookies = QLabel(self.groupBox_login)
        self.label_login_cookies.setObjectName(u"label_login_cookies")

        self.gridLayout_3.addWidget(self.label_login_cookies, 0, 0, 1, 1)

        self.comboBox_browser = QComboBox(self.groupBox_login)
        self.comboBox_browser.setObjectName(u"comboBox_browser")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBox_browser.sizePolicy().hasHeightForWidth())
        self.comboBox_browser.setSizePolicy(sizePolicy)

        self.gridLayout_3.addWidget(self.comboBox_browser, 0, 1, 1, 1)


        self.verticalLayout.addWidget(self.groupBox_login)

        self.groupBox_songfile = QGroupBox(self.tabDownload)
        self.groupBox_songfile.setObjectName(u"groupBox_songfile")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.groupBox_songfile.sizePolicy().hasHeightForWidth())
        self.groupBox_songfile.setSizePolicy(sizePolicy1)
        self.groupBox_songfile.setCheckable(True)
        self.gridLayout_2 = QGridLayout(self.groupBox_songfile)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.label_song_file_encoding = QLabel(self.groupBox_songfile)
        self.label_song_file_encoding.setObjectName(u"label_song_file_encoding")

        self.gridLayout_2.addWidget(self.label_song_file_encoding, 0, 0, 1, 1)

        self.label_song_file_line_endings = QLabel(self.groupBox_songfile)
        self.label_song_file_line_endings.setObjectName(u"label_song_file_line_endings")

        self.gridLayout_2.addWidget(self.label_song_file_line_endings, 1, 0, 1, 1)

        self.comboBox_encoding = QComboBox(self.groupBox_songfile)
        self.comboBox_encoding.setObjectName(u"comboBox_encoding")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(1)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.comboBox_encoding.sizePolicy().hasHeightForWidth())
        self.comboBox_encoding.setSizePolicy(sizePolicy2)

        self.gridLayout_2.addWidget(self.comboBox_encoding, 0, 1, 1, 1)

        self.comboBox_line_endings = QComboBox(self.groupBox_songfile)
        self.comboBox_line_endings.setObjectName(u"comboBox_line_endings")
        sizePolicy2.setHeightForWidth(self.comboBox_line_endings.sizePolicy().hasHeightForWidth())
        self.comboBox_line_endings.setSizePolicy(sizePolicy2)

        self.gridLayout_2.addWidget(self.comboBox_line_endings, 1, 1, 1, 1)

        self.label_song_file_format_version = QLabel(self.groupBox_songfile)
        self.label_song_file_format_version.setObjectName(u"label_song_file_format_version")

        self.gridLayout_2.addWidget(self.label_song_file_format_version, 2, 0, 1, 1)

        self.comboBox_format_version = QComboBox(self.groupBox_songfile)
        self.comboBox_format_version.setObjectName(u"comboBox_format_version")

        self.gridLayout_2.addWidget(self.comboBox_format_version, 2, 1, 1, 1)


        self.verticalLayout.addWidget(self.groupBox_songfile)

        self.groupBox_fixes = QGroupBox(self.tabDownload)
        self.groupBox_fixes.setObjectName(u"groupBox_fixes")
        self.gridLayout_9 = QGridLayout(self.groupBox_fixes)
        self.gridLayout_9.setObjectName(u"gridLayout_9")
        self.label_song_file_fix_linebreaks = QLabel(self.groupBox_fixes)
        self.label_song_file_fix_linebreaks.setObjectName(u"label_song_file_fix_linebreaks")

        self.gridLayout_9.addWidget(self.label_song_file_fix_linebreaks, 0, 0, 1, 1)

        self.comboBox_fix_linebreaks = QComboBox(self.groupBox_fixes)
        self.comboBox_fix_linebreaks.setObjectName(u"comboBox_fix_linebreaks")

        self.gridLayout_9.addWidget(self.comboBox_fix_linebreaks, 0, 1, 1, 1)

        self.label_song_file_fix_first_words_capitalization = QLabel(self.groupBox_fixes)
        self.label_song_file_fix_first_words_capitalization.setObjectName(u"label_song_file_fix_first_words_capitalization")

        self.gridLayout_9.addWidget(self.label_song_file_fix_first_words_capitalization, 1, 0, 1, 1)

        self.checkBox_fix_first_words_capitalization = QCheckBox(self.groupBox_fixes)
        self.checkBox_fix_first_words_capitalization.setObjectName(u"checkBox_fix_first_words_capitalization")

        self.gridLayout_9.addWidget(self.checkBox_fix_first_words_capitalization, 1, 1, 1, 1)

        self.label_song_file_fix_spaces = QLabel(self.groupBox_fixes)
        self.label_song_file_fix_spaces.setObjectName(u"label_song_file_fix_spaces")

        self.gridLayout_9.addWidget(self.label_song_file_fix_spaces, 2, 0, 1, 1)

        self.comboBox_fix_spaces = QComboBox(self.groupBox_fixes)
        self.comboBox_fix_spaces.setObjectName(u"comboBox_fix_spaces")

        self.gridLayout_9.addWidget(self.comboBox_fix_spaces, 2, 1, 1, 1)

        self.label_song_file_fix_quotation_marks = QLabel(self.groupBox_fixes)
        self.label_song_file_fix_quotation_marks.setObjectName(u"label_song_file_fix_quotation_marks")

        self.gridLayout_9.addWidget(self.label_song_file_fix_quotation_marks, 3, 0, 1, 1)

        self.checkBox_fix_quotation_marks = QCheckBox(self.groupBox_fixes)
        self.checkBox_fix_quotation_marks.setObjectName(u"checkBox_fix_quotation_marks")

        self.gridLayout_9.addWidget(self.checkBox_fix_quotation_marks, 3, 1, 1, 1)


        self.verticalLayout.addWidget(self.groupBox_fixes)

        self.groupBox_cover = QGroupBox(self.tabDownload)
        self.groupBox_cover.setObjectName(u"groupBox_cover")
        sizePolicy1.setHeightForWidth(self.groupBox_cover.sizePolicy().hasHeightForWidth())
        self.groupBox_cover.setSizePolicy(sizePolicy1)
        self.groupBox_cover.setCheckable(True)
        self.gridLayout_4 = QGridLayout(self.groupBox_cover)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.label_cover_max_size = QLabel(self.groupBox_cover)
        self.label_cover_max_size.setObjectName(u"label_cover_max_size")

        self.gridLayout_4.addWidget(self.label_cover_max_size, 0, 0, 1, 1)

        self.comboBox_cover_max_size = QComboBox(self.groupBox_cover)
        self.comboBox_cover_max_size.setObjectName(u"comboBox_cover_max_size")

        self.gridLayout_4.addWidget(self.comboBox_cover_max_size, 0, 1, 1, 1)


        self.verticalLayout.addWidget(self.groupBox_cover)

        self.groupBox_background = QGroupBox(self.tabDownload)
        self.groupBox_background.setObjectName(u"groupBox_background")
        sizePolicy1.setHeightForWidth(self.groupBox_background.sizePolicy().hasHeightForWidth())
        self.groupBox_background.setSizePolicy(sizePolicy1)
        self.groupBox_background.setCheckable(True)
        self.gridLayout_8 = QGridLayout(self.groupBox_background)
        self.gridLayout_8.setObjectName(u"gridLayout_8")
        self.checkBox_background_always = QCheckBox(self.groupBox_background)
        self.checkBox_background_always.setObjectName(u"checkBox_background_always")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.checkBox_background_always.sizePolicy().hasHeightForWidth())
        self.checkBox_background_always.setSizePolicy(sizePolicy3)
        self.checkBox_background_always.setChecked(True)

        self.gridLayout_8.addWidget(self.checkBox_background_always, 0, 1, 1, 1)

        self.label_background_always = QLabel(self.groupBox_background)
        self.label_background_always.setObjectName(u"label_background_always")

        self.gridLayout_8.addWidget(self.label_background_always, 0, 0, 1, 1)


        self.verticalLayout.addWidget(self.groupBox_background)


        self.horizontalLayout.addLayout(self.verticalLayout)

        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.groupBox_2 = QGroupBox(self.tabDownload)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.gridLayout_10 = QGridLayout(self.groupBox_2)
        self.gridLayout_10.setObjectName(u"gridLayout_10")
        self.label_ytdlp_rate_limit = QLabel(self.groupBox_2)
        self.label_ytdlp_rate_limit.setObjectName(u"label_ytdlp_rate_limit")

        self.gridLayout_10.addWidget(self.label_ytdlp_rate_limit, 1, 0, 1, 1)

        self.comboBox_ytdlp_rate_limit = QComboBox(self.groupBox_2)
        self.comboBox_ytdlp_rate_limit.setObjectName(u"comboBox_ytdlp_rate_limit")

        self.gridLayout_10.addWidget(self.comboBox_ytdlp_rate_limit, 1, 1, 1, 1)

        self.label_throttling_threads = QLabel(self.groupBox_2)
        self.label_throttling_threads.setObjectName(u"label_throttling_threads")

        self.gridLayout_10.addWidget(self.label_throttling_threads, 0, 0, 1, 1)

        self.spinBox_throttling_threads = QSpinBox(self.groupBox_2)
        self.spinBox_throttling_threads.setObjectName(u"spinBox_throttling_threads")
        self.spinBox_throttling_threads.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.spinBox_throttling_threads.setMinimum(0)
        self.spinBox_throttling_threads.setMaximum(32)

        self.gridLayout_10.addWidget(self.spinBox_throttling_threads, 0, 1, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_2)

        self.groupBox_audio = QGroupBox(self.tabDownload)
        self.groupBox_audio.setObjectName(u"groupBox_audio")
        sizePolicy1.setHeightForWidth(self.groupBox_audio.sizePolicy().hasHeightForWidth())
        self.groupBox_audio.setSizePolicy(sizePolicy1)
        self.groupBox_audio.setCheckable(True)
        self.gridLayout_5 = QGridLayout(self.groupBox_audio)
        self.gridLayout_5.setObjectName(u"gridLayout_5")
        self.checkBox_audio_embed_artwork = QCheckBox(self.groupBox_audio)
        self.checkBox_audio_embed_artwork.setObjectName(u"checkBox_audio_embed_artwork")

        self.gridLayout_5.addWidget(self.checkBox_audio_embed_artwork, 3, 3, 1, 1)

        self.label_audio_embed_artwork = QLabel(self.groupBox_audio)
        self.label_audio_embed_artwork.setObjectName(u"label_audio_embed_artwork")

        self.gridLayout_5.addWidget(self.label_audio_embed_artwork, 3, 0, 1, 1)

        self.label_audio_bitrate = QLabel(self.groupBox_audio)
        self.label_audio_bitrate.setObjectName(u"label_audio_bitrate")

        self.gridLayout_5.addWidget(self.label_audio_bitrate, 1, 0, 1, 1)

        self.comboBox_audio_bitrate = QComboBox(self.groupBox_audio)
        self.comboBox_audio_bitrate.setObjectName(u"comboBox_audio_bitrate")

        self.gridLayout_5.addWidget(self.comboBox_audio_bitrate, 1, 3, 1, 1)

        self.label_audio_normalize = QLabel(self.groupBox_audio)
        self.label_audio_normalize.setObjectName(u"label_audio_normalize")

        self.gridLayout_5.addWidget(self.label_audio_normalize, 2, 0, 1, 1)

        self.label_audio_format = QLabel(self.groupBox_audio)
        self.label_audio_format.setObjectName(u"label_audio_format")

        self.gridLayout_5.addWidget(self.label_audio_format, 0, 0, 1, 1)

        self.comboBox_audio_format = QComboBox(self.groupBox_audio)
        self.comboBox_audio_format.setObjectName(u"comboBox_audio_format")
        sizePolicy.setHeightForWidth(self.comboBox_audio_format.sizePolicy().hasHeightForWidth())
        self.comboBox_audio_format.setSizePolicy(sizePolicy)

        self.gridLayout_5.addWidget(self.comboBox_audio_format, 0, 3, 1, 1)

        self.comboBox_audio_normalization = QComboBox(self.groupBox_audio)
        self.comboBox_audio_normalization.setObjectName(u"comboBox_audio_normalization")

        self.gridLayout_5.addWidget(self.comboBox_audio_normalization, 2, 3, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_audio)

        self.groupBox_video = QGroupBox(self.tabDownload)
        self.groupBox_video.setObjectName(u"groupBox_video")
        sizePolicy1.setHeightForWidth(self.groupBox_video.sizePolicy().hasHeightForWidth())
        self.groupBox_video.setSizePolicy(sizePolicy1)
        self.groupBox_video.setCheckable(True)
        self.gridLayout_6 = QGridLayout(self.groupBox_video)
        self.gridLayout_6.setObjectName(u"gridLayout_6")
        self.groupBox_reencode_video = QGroupBox(self.groupBox_video)
        self.groupBox_reencode_video.setObjectName(u"groupBox_reencode_video")
        sizePolicy1.setHeightForWidth(self.groupBox_reencode_video.sizePolicy().hasHeightForWidth())
        self.groupBox_reencode_video.setSizePolicy(sizePolicy1)
        self.groupBox_reencode_video.setCheckable(True)
        self.groupBox_reencode_video.setChecked(False)
        self.gridLayout_7 = QGridLayout(self.groupBox_reencode_video)
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.label_18 = QLabel(self.groupBox_reencode_video)
        self.label_18.setObjectName(u"label_18")

        self.gridLayout_7.addWidget(self.label_18, 0, 0, 1, 1)

        self.comboBox_videoencoder = QComboBox(self.groupBox_reencode_video)
        self.comboBox_videoencoder.setObjectName(u"comboBox_videoencoder")
        sizePolicy2.setHeightForWidth(self.comboBox_videoencoder.sizePolicy().hasHeightForWidth())
        self.comboBox_videoencoder.setSizePolicy(sizePolicy2)

        self.gridLayout_7.addWidget(self.comboBox_videoencoder, 0, 1, 1, 1)


        self.gridLayout_6.addWidget(self.groupBox_reencode_video, 4, 0, 1, 2)

        self.comboBox_fps = QComboBox(self.groupBox_video)
        self.comboBox_fps.setObjectName(u"comboBox_fps")
        sizePolicy2.setHeightForWidth(self.comboBox_fps.sizePolicy().hasHeightForWidth())
        self.comboBox_fps.setSizePolicy(sizePolicy2)

        self.gridLayout_6.addWidget(self.comboBox_fps, 2, 1, 1, 1)

        self.checkBox_video_embed_artwork = QCheckBox(self.groupBox_video)
        self.checkBox_video_embed_artwork.setObjectName(u"checkBox_video_embed_artwork")

        self.gridLayout_6.addWidget(self.checkBox_video_embed_artwork, 3, 1, 1, 1)

        self.label_video_container = QLabel(self.groupBox_video)
        self.label_video_container.setObjectName(u"label_video_container")

        self.gridLayout_6.addWidget(self.label_video_container, 0, 0, 1, 1)

        self.comboBox_videocontainer = QComboBox(self.groupBox_video)
        self.comboBox_videocontainer.setObjectName(u"comboBox_videocontainer")
        sizePolicy2.setHeightForWidth(self.comboBox_videocontainer.sizePolicy().hasHeightForWidth())
        self.comboBox_videocontainer.setSizePolicy(sizePolicy2)

        self.gridLayout_6.addWidget(self.comboBox_videocontainer, 0, 1, 1, 1)

        self.label_video_max_fps = QLabel(self.groupBox_video)
        self.label_video_max_fps.setObjectName(u"label_video_max_fps")

        self.gridLayout_6.addWidget(self.label_video_max_fps, 2, 0, 1, 1)

        self.label_video_embed_artwork = QLabel(self.groupBox_video)
        self.label_video_embed_artwork.setObjectName(u"label_video_embed_artwork")

        self.gridLayout_6.addWidget(self.label_video_embed_artwork, 3, 0, 1, 1)

        self.label_video_max_resolution = QLabel(self.groupBox_video)
        self.label_video_max_resolution.setObjectName(u"label_video_max_resolution")

        self.gridLayout_6.addWidget(self.label_video_max_resolution, 1, 0, 1, 1)

        self.comboBox_videoresolution = QComboBox(self.groupBox_video)
        self.comboBox_videoresolution.setObjectName(u"comboBox_videoresolution")
        sizePolicy2.setHeightForWidth(self.comboBox_videoresolution.sizePolicy().hasHeightForWidth())
        self.comboBox_videoresolution.setSizePolicy(sizePolicy2)

        self.gridLayout_6.addWidget(self.comboBox_videoresolution, 1, 1, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_video)

        self.groupBox_3 = QGroupBox(self.tabDownload)
        self.groupBox_3.setObjectName(u"groupBox_3")
        self.groupBox_3.setMinimumSize(QSize(0, 80))
        self.gridLayout_11 = QGridLayout(self.groupBox_3)
        self.gridLayout_11.setObjectName(u"gridLayout_11")
        self.label_discord = QLabel(self.groupBox_3)
        self.label_discord.setObjectName(u"label_discord")

        self.gridLayout_11.addWidget(self.label_discord, 0, 0, 1, 1)

        self.checkBox_discord_allowed = QCheckBox(self.groupBox_3)
        self.checkBox_discord_allowed.setObjectName(u"checkBox_discord_allowed")

        self.gridLayout_11.addWidget(self.checkBox_discord_allowed, 0, 1, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox_3)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)


        self.horizontalLayout.addLayout(self.verticalLayout_2)

        self.horizontalLayout.setStretch(0, 1)
        self.horizontalLayout.setStretch(1, 1)
        self.tabWidget.addTab(self.tabDownload, "")
        self.tabFiles = QWidget()
        self.tabFiles.setObjectName(u"tabFiles")
        self.verticalLayout_4 = QVBoxLayout(self.tabFiles)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.groupBox = QGroupBox(self.tabFiles)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_3 = QVBoxLayout(self.groupBox)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.edit_path_template = QLineEdit(self.groupBox)
        self.edit_path_template.setObjectName(u"edit_path_template")
        self.edit_path_template.setMaxLength(100)

        self.horizontalLayout_3.addWidget(self.edit_path_template)

        self.button_default_path_template = QPushButton(self.groupBox)
        self.button_default_path_template.setObjectName(u"button_default_path_template")

        self.horizontalLayout_3.addWidget(self.button_default_path_template)


        self.verticalLayout_3.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.comboBox_placeholder = QComboBox(self.groupBox)
        self.comboBox_placeholder.setObjectName(u"comboBox_placeholder")

        self.horizontalLayout_2.addWidget(self.comboBox_placeholder)

        self.button_insert_placeholder = QPushButton(self.groupBox)
        self.button_insert_placeholder.setObjectName(u"button_insert_placeholder")

        self.horizontalLayout_2.addWidget(self.button_insert_placeholder)


        self.verticalLayout_3.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label = QLabel(self.groupBox)
        self.label.setObjectName(u"label")

        self.horizontalLayout_4.addWidget(self.label)

        self.edit_path_template_result = QLineEdit(self.groupBox)
        self.edit_path_template_result.setObjectName(u"edit_path_template_result")
        self.edit_path_template_result.setReadOnly(True)

        self.horizontalLayout_4.addWidget(self.edit_path_template_result)


        self.verticalLayout_3.addLayout(self.horizontalLayout_4)


        self.verticalLayout_4.addWidget(self.groupBox)

        self.verticalSpacer_3 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_4.addItem(self.verticalSpacer_3)

        self.tabWidget.addTab(self.tabFiles, "")
        self.tabPaths = QWidget()
        self.tabPaths.setObjectName(u"tabPaths")
        self.verticalLayout_5 = QVBoxLayout(self.tabPaths)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.groupBox_usdx = QGroupBox(self.tabPaths)
        self.groupBox_usdx.setObjectName(u"groupBox_usdx")
        self.horizontalLayout_5 = QHBoxLayout(self.groupBox_usdx)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(6, 6, 6, 6)
        self.label_usdx = QLabel(self.groupBox_usdx)
        self.label_usdx.setObjectName(u"label_usdx")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.label_usdx.sizePolicy().hasHeightForWidth())
        self.label_usdx.setSizePolicy(sizePolicy4)
        self.label_usdx.setMinimumSize(QSize(32, 32))
        self.label_usdx.setMaximumSize(QSize(32, 32))
        self.label_usdx.setPixmap(QPixmap(u":/icons/usdx.png"))
        self.label_usdx.setScaledContents(True)

        self.horizontalLayout_5.addWidget(self.label_usdx)

        self.lineEdit_path_usdx = QLineEdit(self.groupBox_usdx)
        self.lineEdit_path_usdx.setObjectName(u"lineEdit_path_usdx")
        self.lineEdit_path_usdx.setReadOnly(True)

        self.horizontalLayout_5.addWidget(self.lineEdit_path_usdx)

        self.pushButton_browse_usdx = QPushButton(self.groupBox_usdx)
        self.pushButton_browse_usdx.setObjectName(u"pushButton_browse_usdx")

        self.horizontalLayout_5.addWidget(self.pushButton_browse_usdx)


        self.verticalLayout_5.addWidget(self.groupBox_usdx)

        self.groupBox_vocaluxe = QGroupBox(self.tabPaths)
        self.groupBox_vocaluxe.setObjectName(u"groupBox_vocaluxe")
        self.horizontalLayout_8 = QHBoxLayout(self.groupBox_vocaluxe)
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.horizontalLayout_8.setContentsMargins(6, 6, 6, 6)
        self.label_vocaluxe = QLabel(self.groupBox_vocaluxe)
        self.label_vocaluxe.setObjectName(u"label_vocaluxe")
        sizePolicy4.setHeightForWidth(self.label_vocaluxe.sizePolicy().hasHeightForWidth())
        self.label_vocaluxe.setSizePolicy(sizePolicy4)
        self.label_vocaluxe.setMinimumSize(QSize(32, 32))
        self.label_vocaluxe.setMaximumSize(QSize(32, 32))
        self.label_vocaluxe.setPixmap(QPixmap(u":/icons/vocaluxe.png"))
        self.label_vocaluxe.setScaledContents(True)

        self.horizontalLayout_8.addWidget(self.label_vocaluxe)

        self.lineEdit_path_vocaluxe = QLineEdit(self.groupBox_vocaluxe)
        self.lineEdit_path_vocaluxe.setObjectName(u"lineEdit_path_vocaluxe")
        self.lineEdit_path_vocaluxe.setReadOnly(True)

        self.horizontalLayout_8.addWidget(self.lineEdit_path_vocaluxe)

        self.pushButton_browse_vocaluxe = QPushButton(self.groupBox_vocaluxe)
        self.pushButton_browse_vocaluxe.setObjectName(u"pushButton_browse_vocaluxe")

        self.horizontalLayout_8.addWidget(self.pushButton_browse_vocaluxe)


        self.verticalLayout_5.addWidget(self.groupBox_vocaluxe)

        self.groupBox_performous = QGroupBox(self.tabPaths)
        self.groupBox_performous.setObjectName(u"groupBox_performous")
        self.horizontalLayout_10 = QHBoxLayout(self.groupBox_performous)
        self.horizontalLayout_10.setObjectName(u"horizontalLayout_10")
        self.horizontalLayout_10.setContentsMargins(6, 6, 6, 6)
        self.label_performous = QLabel(self.groupBox_performous)
        self.label_performous.setObjectName(u"label_performous")
        sizePolicy4.setHeightForWidth(self.label_performous.sizePolicy().hasHeightForWidth())
        self.label_performous.setSizePolicy(sizePolicy4)
        self.label_performous.setMinimumSize(QSize(32, 32))
        self.label_performous.setMaximumSize(QSize(32, 32))
        self.label_performous.setPixmap(QPixmap(u":/icons/performous.png"))
        self.label_performous.setScaledContents(True)

        self.horizontalLayout_10.addWidget(self.label_performous)

        self.lineEdit_path_performous = QLineEdit(self.groupBox_performous)
        self.lineEdit_path_performous.setObjectName(u"lineEdit_path_performous")
        self.lineEdit_path_performous.setReadOnly(True)

        self.horizontalLayout_10.addWidget(self.lineEdit_path_performous)

        self.pushButton_browse_performous = QPushButton(self.groupBox_performous)
        self.pushButton_browse_performous.setObjectName(u"pushButton_browse_performous")

        self.horizontalLayout_10.addWidget(self.pushButton_browse_performous)


        self.verticalLayout_5.addWidget(self.groupBox_performous)

        self.groupBox_ultrastar_manager = QGroupBox(self.tabPaths)
        self.groupBox_ultrastar_manager.setObjectName(u"groupBox_ultrastar_manager")
        self.horizontalLayout_9 = QHBoxLayout(self.groupBox_ultrastar_manager)
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")
        self.horizontalLayout_9.setContentsMargins(6, 6, 6, 6)
        self.label_ultrastar_manager = QLabel(self.groupBox_ultrastar_manager)
        self.label_ultrastar_manager.setObjectName(u"label_ultrastar_manager")
        sizePolicy4.setHeightForWidth(self.label_ultrastar_manager.sizePolicy().hasHeightForWidth())
        self.label_ultrastar_manager.setSizePolicy(sizePolicy4)
        self.label_ultrastar_manager.setMinimumSize(QSize(32, 32))
        self.label_ultrastar_manager.setMaximumSize(QSize(32, 32))
        self.label_ultrastar_manager.setPixmap(QPixmap(u":/icons/ultrastar-manager.png"))
        self.label_ultrastar_manager.setScaledContents(True)

        self.horizontalLayout_9.addWidget(self.label_ultrastar_manager)

        self.lineEdit_path_ultrastar_manager = QLineEdit(self.groupBox_ultrastar_manager)
        self.lineEdit_path_ultrastar_manager.setObjectName(u"lineEdit_path_ultrastar_manager")
        self.lineEdit_path_ultrastar_manager.setReadOnly(True)

        self.horizontalLayout_9.addWidget(self.lineEdit_path_ultrastar_manager)

        self.pushButton_browse_ultrastar_manager = QPushButton(self.groupBox_ultrastar_manager)
        self.pushButton_browse_ultrastar_manager.setObjectName(u"pushButton_browse_ultrastar_manager")

        self.horizontalLayout_9.addWidget(self.pushButton_browse_ultrastar_manager)


        self.verticalLayout_5.addWidget(self.groupBox_ultrastar_manager)

        self.groupBox_yass_reloaded = QGroupBox(self.tabPaths)
        self.groupBox_yass_reloaded.setObjectName(u"groupBox_yass_reloaded")
        self.horizontalLayout_6 = QHBoxLayout(self.groupBox_yass_reloaded)
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.horizontalLayout_6.setContentsMargins(6, 6, 6, 6)
        self.label_yass_reloaded = QLabel(self.groupBox_yass_reloaded)
        self.label_yass_reloaded.setObjectName(u"label_yass_reloaded")
        sizePolicy4.setHeightForWidth(self.label_yass_reloaded.sizePolicy().hasHeightForWidth())
        self.label_yass_reloaded.setSizePolicy(sizePolicy4)
        self.label_yass_reloaded.setMinimumSize(QSize(32, 32))
        self.label_yass_reloaded.setMaximumSize(QSize(32, 32))
        self.label_yass_reloaded.setPixmap(QPixmap(u":/icons/yass-reloaded.png"))
        self.label_yass_reloaded.setScaledContents(True)

        self.horizontalLayout_6.addWidget(self.label_yass_reloaded)

        self.lineEdit_path_yass_reloaded = QLineEdit(self.groupBox_yass_reloaded)
        self.lineEdit_path_yass_reloaded.setObjectName(u"lineEdit_path_yass_reloaded")
        self.lineEdit_path_yass_reloaded.setReadOnly(True)

        self.horizontalLayout_6.addWidget(self.lineEdit_path_yass_reloaded)

        self.pushButton_browse_yass_reloaded = QPushButton(self.groupBox_yass_reloaded)
        self.pushButton_browse_yass_reloaded.setObjectName(u"pushButton_browse_yass_reloaded")

        self.horizontalLayout_6.addWidget(self.pushButton_browse_yass_reloaded)


        self.verticalLayout_5.addWidget(self.groupBox_yass_reloaded)

        self.groupBox_karedi = QGroupBox(self.tabPaths)
        self.groupBox_karedi.setObjectName(u"groupBox_karedi")
        self.horizontalLayout_7 = QHBoxLayout(self.groupBox_karedi)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.horizontalLayout_7.setContentsMargins(6, 6, 6, 6)
        self.label_karedi = QLabel(self.groupBox_karedi)
        self.label_karedi.setObjectName(u"label_karedi")
        sizePolicy4.setHeightForWidth(self.label_karedi.sizePolicy().hasHeightForWidth())
        self.label_karedi.setSizePolicy(sizePolicy4)
        self.label_karedi.setMinimumSize(QSize(32, 32))
        self.label_karedi.setMaximumSize(QSize(32, 32))
        self.label_karedi.setPixmap(QPixmap(u":/icons/karedi.png"))
        self.label_karedi.setScaledContents(True)

        self.horizontalLayout_7.addWidget(self.label_karedi)

        self.lineEdit_path_karedi = QLineEdit(self.groupBox_karedi)
        self.lineEdit_path_karedi.setObjectName(u"lineEdit_path_karedi")
        self.lineEdit_path_karedi.setReadOnly(True)

        self.horizontalLayout_7.addWidget(self.lineEdit_path_karedi)

        self.pushButton_browse_karedi = QPushButton(self.groupBox_karedi)
        self.pushButton_browse_karedi.setObjectName(u"pushButton_browse_karedi")

        self.horizontalLayout_7.addWidget(self.pushButton_browse_karedi)


        self.verticalLayout_5.addWidget(self.groupBox_karedi)

        self.tabWidget.addTab(self.tabPaths, "")

        self.gridLayout.addWidget(self.tabWidget, 0, 0, 2, 2)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_2, 2, 0, 1, 1)


        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Settings", None))
        self.label_theme.setText(QCoreApplication.translate("Dialog", u"Theme:", None))
        self.label_primary_color.setText(QCoreApplication.translate("Dialog", u"Primary color:", None))
        self.label_3.setText("")
        self.checkBox_colored_background.setText(QCoreApplication.translate("Dialog", u"Colored background", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QCoreApplication.translate("Dialog", u"View", None))
        self.groupBox_login.setTitle(QCoreApplication.translate("Dialog", u"Login cookies (USDB and Youtube)", None))
#if QT_CONFIG(tooltip)
        self.label_login_cookies.setToolTip("")
#endif // QT_CONFIG(tooltip)
        self.label_login_cookies.setText(QCoreApplication.translate("Dialog", u"Browser:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_browser.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Age-restricted videos require a login. Choose the browser you are logged in with, so that the respective cookie can be retrieved automatically.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.groupBox_songfile.setTitle(QCoreApplication.translate("Dialog", u"Download song file", None))
#if QT_CONFIG(tooltip)
        self.label_song_file_encoding.setToolTip("")
#endif // QT_CONFIG(tooltip)
        self.label_song_file_encoding.setText(QCoreApplication.translate("Dialog", u"Encoding:", None))
        self.label_song_file_line_endings.setText(QCoreApplication.translate("Dialog", u"Line endings:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_encoding.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Choose UTF-8 for USDX, US Play, US Worldparty and Performous.</p><p>Choose UTF-8 BOM for Vocaluxe (Byte Order Mark may cause trouble in some text editors).</p><p>Choose CP1252 for USDX (pre 1.1) and USDX CMD. </p></body></html>", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.comboBox_line_endings.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>USDX and variants can handle both line ending types, but if you want to edit the song file manually, your editor of choice may require a certain line ending. UNIX users (MacOS, Linux) should generally use LF (line feed). On Windows, some text editors may require CRLF (carriage return, line feed). USDB currently requires CRLF line endings.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_song_file_format_version.setText(QCoreApplication.translate("Dialog", u"Version:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_format_version.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p><span style=\" font-weight:700;\">V1.0.0</span></p><p>Choose this version for highest compatibility. This uses #MP3 to specifiy the audio file.</p><p><span style=\" font-weight:700;\">V1.1.0</span></p><p>This version switches from #MP3 to #AUDIO to underline that other audio formats than mp3 are supported. For compatibility, the Syncer adds both #MP3 and #AUDIO to the text file, which should cause no issues when loading the song.</p><p><span style=\" font-weight:700;\">V1.2.0</span></p><p>This version introduces new tags #AUDIOURL, #COVERURL, #BACKGROUNDURL and #VIDEOURL to specify URLs for the respective sources. The syncer will fill these tags with the contents of the respective metatags, if available.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.groupBox_fixes.setTitle(QCoreApplication.translate("Dialog", u"Optional song file fixes", None))
        self.label_song_file_fix_linebreaks.setText(QCoreApplication.translate("Dialog", u"Fix linebreaks:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_fix_linebreaks.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>If this setting is enabled, the linebreak timings will be recalculated, either according to USDX or to YASS rules.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_song_file_fix_first_words_capitalization.setText(QCoreApplication.translate("Dialog", u"Fix capitalization:", None))
        self.checkBox_fix_first_words_capitalization.setText("")
        self.label_song_file_fix_spaces.setText(QCoreApplication.translate("Dialog", u"Fix spaces:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_fix_spaces.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>If this setting is enabled, all inter-word spaces will be shifted to either after or before a word.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_song_file_fix_quotation_marks.setText(QCoreApplication.translate("Dialog", u"Fix quotation marks:", None))
        self.checkBox_fix_quotation_marks.setText("")
        self.groupBox_cover.setTitle(QCoreApplication.translate("Dialog", u"Download cover", None))
        self.label_cover_max_size.setText(QCoreApplication.translate("Dialog", u"Max. size:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_cover_max_size.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Images larger than this will be scaled to the set value.</p><p>Some karaoke software variants may have problems displaying covers larger than 1920x1920 pixels.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.groupBox_background.setTitle(QCoreApplication.translate("Dialog", u"Download background", None))
#if QT_CONFIG(tooltip)
        self.checkBox_background_always.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>If checked, background images will be downloaded even if there is a video for the song as well.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.checkBox_background_always.setText("")
        self.label_background_always.setText(QCoreApplication.translate("Dialog", u"Even with video:", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("Dialog", u"Throttling", None))
        self.label_ytdlp_rate_limit.setText(QCoreApplication.translate("Dialog", u"YouTube rate limit:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_ytdlp_rate_limit.setToolTip(QCoreApplication.translate("Dialog", u"Limit YouTube download bandwidth to avoid getting blocked (per thread)", None))
#endif // QT_CONFIG(tooltip)
        self.label_throttling_threads.setText(QCoreApplication.translate("Dialog", u"Download threads:", None))
#if QT_CONFIG(tooltip)
        self.spinBox_throttling_threads.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>0: use default (number of cores)</p><p>Requires restart to take effect.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.groupBox_audio.setTitle(QCoreApplication.translate("Dialog", u"Download audio", None))
        self.checkBox_audio_embed_artwork.setText("")
        self.label_audio_embed_artwork.setText(QCoreApplication.translate("Dialog", u"Embed artwork:", None))
        self.label_audio_bitrate.setText(QCoreApplication.translate("Dialog", u"Bitrate:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_audio_bitrate.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>If the audio file is reencoded, the selected bitrate will be used.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_audio_normalize.setText(QCoreApplication.translate("Dialog", u"Normalization:", None))
        self.label_audio_format.setText(QCoreApplication.translate("Dialog", u"Format:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_audio_format.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p><span style=\" font-weight:700;\">M4A:</span> Best quality and speed. Compatible with UltraStar Deluxe, Vocaluxe, Performous and UltraStar World Party.</p><p><span style=\" font-weight:700;\">MP3: </span>Worst quality and speed. Full compatibility.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.comboBox_audio_normalization.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>It is recommended to adjust the volume of your audio files so that they all have the same level of perceived loudness.</p><p>If you use a recent version of USDX (2025.4.0 or newer), then choose <span style=\" font-weight:700;\">ReplayGain</span>, as this does not reencode the file but instead writes the normalization information into the audio header. You have to enable ReplayGain in USDX sound settings for this take effect.</p><p>For karaoke softwares that do not (yet) support replay gain, choose <span style=\" font-weight:700;\">Normalize</span>. This will adjust the loudness by re-encoding the file.</p><p>To skip normalization, choose <span style=\" font-weight:700;\">disabled</span>. This saves time, but resulting audio files will have different loudness levels.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.groupBox_video.setTitle(QCoreApplication.translate("Dialog", u"Download video", None))
#if QT_CONFIG(tooltip)
        self.groupBox_reencode_video.setToolTip("")
#endif // QT_CONFIG(tooltip)
        self.groupBox_reencode_video.setTitle(QCoreApplication.translate("Dialog", u"Reencode video", None))
        self.label_18.setText(QCoreApplication.translate("Dialog", u"New format:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_videoencoder.setToolTip("")
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.comboBox_fps.setToolTip(QCoreApplication.translate("Dialog", u"Maximum frames per second. Lower values reduce file size.", None))
#endif // QT_CONFIG(tooltip)
        self.checkBox_video_embed_artwork.setText("")
        self.label_video_container.setText(QCoreApplication.translate("Dialog", u"Container/codec:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_videocontainer.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>For <span style=\" font-weight:700;\">UltraStar deluxe</span>, <span style=\" font-weight:700;\">Performous</span>, <span style=\" font-weight:700;\">Vocaluxe</span> or other ffmpeg-based karaoke softwares, choose &quot;Best available container/codec&quot; or &quot;mp4 (best available codec)&quot; if you prefer to have an mp4 container.</p><p>For <span style=\" font-weight:700;\">UltraStar CMD</span> (Windows), <span style=\" font-weight:700;\">Melody Mania</span> (Windows) and low-power computers, choose &quot;mp4 (AVC/H.264).</p><p>Otherwise, select a container/codec option that your software supports.</p><p>If the selected container/format option is not available for a given song, the syncer will try to at least get the requested container, and if that is not available either, it will fallback to the best available container/codec.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_video_max_fps.setText(QCoreApplication.translate("Dialog", u"Max. FPS:", None))
        self.label_video_embed_artwork.setText(QCoreApplication.translate("Dialog", u"Embed artwork:", None))
        self.label_video_max_resolution.setText(QCoreApplication.translate("Dialog", u"Max. resolution:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_videoresolution.setToolTip(QCoreApplication.translate("Dialog", u"Maximum video height. Lower values reduce file size.", None))
#endif // QT_CONFIG(tooltip)
        self.groupBox_3.setTitle(QCoreApplication.translate("Dialog", u"Discord integration for outdated resources", None))
        self.label_discord.setText(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Allow <span style=\" font-style:italic;\">anonymous</span> notifications:</p></body></html>", None))
#if QT_CONFIG(tooltip)
        self.checkBox_discord_allowed.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Enabling this will allow reporting unavailable resource \u2013 <span style=\" font-weight:700;\">entirely anonymously</span> \u2013 to our Karaoke Discord channel so that they can be fixed quickly. Please enable this to contribute to the community.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.checkBox_discord_allowed.setText("")
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabDownload), QCoreApplication.translate("Dialog", u"Download", None))
#if QT_CONFIG(tooltip)
        self.groupBox.setToolTip(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Determines how song files are stored relative to the song directory. You can use certain placeholder names to reference song attributes.</p><p>The template must contain at least two components, which are separated using slashes. The last component specifies the filename, excluding its extension.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.groupBox.setTitle(QCoreApplication.translate("Dialog", u"Path template", None))
        self.button_default_path_template.setText(QCoreApplication.translate("Dialog", u"Default", None))
        self.button_insert_placeholder.setText(QCoreApplication.translate("Dialog", u"Insert placeholder", None))
        self.label.setText(QCoreApplication.translate("Dialog", u"Result:", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabFiles), QCoreApplication.translate("Dialog", u"Files", None))
        self.groupBox_usdx.setTitle(QCoreApplication.translate("Dialog", u"UltraStar deluxe", None))
        self.label_usdx.setText("")
        self.pushButton_browse_usdx.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.groupBox_vocaluxe.setTitle(QCoreApplication.translate("Dialog", u"Vocaluxe", None))
        self.label_vocaluxe.setText("")
        self.pushButton_browse_vocaluxe.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.groupBox_performous.setTitle(QCoreApplication.translate("Dialog", u"Performous", None))
        self.label_performous.setText("")
        self.pushButton_browse_performous.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.groupBox_ultrastar_manager.setTitle(QCoreApplication.translate("Dialog", u"UltraStar Manager", None))
        self.label_ultrastar_manager.setText("")
        self.pushButton_browse_ultrastar_manager.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.groupBox_yass_reloaded.setTitle(QCoreApplication.translate("Dialog", u"YASS reloaded", None))
        self.label_yass_reloaded.setText("")
        self.pushButton_browse_yass_reloaded.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.groupBox_karedi.setTitle(QCoreApplication.translate("Dialog", u"Karedi", None))
        self.label_karedi.setText("")
        self.pushButton_browse_karedi.setText(QCoreApplication.translate("Dialog", u"...", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabPaths), QCoreApplication.translate("Dialog", u"App Paths", None))
    # retranslateUi




"""Helper classes and functions related to the song text file"""

# Since this file deals with a bunch of ambiguous unicode characters,
# we disable the rule file-wide.
# ruff: noqa: RUF001, RUF003

import math
from typing import NamedTuple

import attrs

from usdb_syncer.constants import (
    LANGUAGES_WITH_SPACED_QUOTES,
    MINIMUM_BPM,
    QUOTATION_MARKS,
    QUOTATION_MARKS_TO_REPLACE,
)


class QuotationMarkReplacementResult(NamedTuple):
    """Named tuple for the result of replace_false_quotation_marks."""

    text: str
    marks_fixed: int
    opening: bool


@attrs.define
class BeatsPerMinute:
    """New type for beats per minute float."""

    value: float = NotImplemented

    def __str__(self) -> str:
        return f"{round(self.value, 2):g}"

    @classmethod
    def parse(cls, value: str) -> "BeatsPerMinute":
        return cls(float(value.replace(",", ".")))

    def beats_to_secs(self, beats: int) -> float:
        return beats / (self.value * 4) * 60

    def secs_to_beats(self, secs: float) -> int:
        return int(secs * 4 * self.value / 60)

    def beats_to_ms(self, beats: int) -> float:
        return self.beats_to_secs(beats) * 1000

    def is_too_low(self) -> bool:
        return self.value < MINIMUM_BPM

    def make_large_enough(self) -> int:
        """Double BPM (if necessary, multiple times) until it is above MINIMUM_BPM
        and returns the required multiplication factor."""
        # how often to double bpm until it is larger or equal to the threshold
        exp = math.ceil(math.log2(MINIMUM_BPM / self.value))
        factor = 2**exp
        self.value = self.value * factor
        return factor


def replace_false_apostrophes(value: str) -> str:
    # two single upright quotation marks ('') by double upright quotation marks (")
    # grave (`) and acute accent (´), prime symbol (′), left single quotation mark (‘)
    # and upright apostrophe (') by typographer’s apostrophe (’)
    return (
        value.replace("''", '"')
        .replace("`", "’")
        .replace("´", "’")
        .replace("′", "’")
        .replace("‘", "’")
        .replace("'", "’")
    )


def replace_false_quotation_marks(
    text: str, language: str | None, opening: bool
) -> QuotationMarkReplacementResult:
    # replaces quotation marks with the correct, language-specific ones
    # Note: nested quotation marks as in "Hello “world”" are not supported
    if language:
        opening_quote, closing_quote = QUOTATION_MARKS.get(language, ("“", "”"))
    else:
        opening_quote, closing_quote = ("“", "”")

    spaced_quotes = language in LANGUAGES_WITH_SPACED_QUOTES
    new_text = []
    marks_fixed = 0

    i = 0
    while i < len(text):
        char = text[i]
        if char in QUOTATION_MARKS_TO_REPLACE:
            if opening:
                new_text.append(opening_quote)
                if spaced_quotes:
                    new_text.append(" ")
                    if i < len(text) - 1 and text[i + 1].isspace():
                        i += 1  # skip space
            else:
                if spaced_quotes:
                    if i > 0 and text[i - 1].isspace():
                        new_text.pop()  # remove already appended whitespace
                    new_text.append(" ")
                new_text.append(closing_quote)
            opening = not opening
            marks_fixed += 1
        else:
            new_text.append(char)
        i += 1

    text = "".join(new_text)

    return QuotationMarkReplacementResult(text, marks_fixed, opening)



"""Options for downloading songs."""

from dataclasses import dataclass
from pathlib import Path

from usdb_syncer import path_template, settings


@dataclass(frozen=True)
class TxtOptions:
    """Settings regarding the song txt file to be downloaded."""

    encoding: settings.Encoding
    newline: settings.Newline
    format_version: settings.FormatVersion
    fix_linebreaks: settings.FixLinebreaks
    fix_first_words_capitalization: bool
    fix_spaces: settings.FixSpaces
    fix_quotation_marks: bool


@dataclass(frozen=True)
class AudioOptions:
    """Settings regarding the audio file to be downloaded."""

    format: settings.AudioFormat
    bitrate: settings.AudioBitrate
    normalization: settings.AudioNormalization
    embed_artwork: bool
    rate_limit: settings.YtdlpRateLimit

    def ytdl_format(self) -> str:
        return self.format.ytdl_format()


@dataclass(frozen=True)
class VideoOptions:
    """Settings regarding the video file to be downloaded."""

    format: settings.VideoContainer
    reencode_format: settings.VideoCodec | None
    max_resolution: settings.VideoResolution
    max_fps: settings.VideoFps
    embed_artwork: bool
    rate_limit: settings.YtdlpRateLimit

    def ytdl_format(self) -> str:
        container = f"[ext={self.format.ytdl_ext()}]" if self.format.ytdl_ext() else ""
        codec = (
            f"[vcodec~='{self.format.ytdl_vcodec()}']"
            if self.format.ytdl_vcodec()
            else ""
        )
        fmt = ["bestvideo*"]
        if container:
            fmt.insert(0, f"bestvideo*{container}")
        if codec:
            fmt.insert(0, f"bestvideo*{container}{codec}")
        width = f"[width<={self.max_resolution.width()}]"
        height = f"[height<={self.max_resolution.height()}]"
        fps = f"[fps<={self.max_fps.value}]"
        # fps filter always fails for some platforms, so skip it as a fallback
        # also some formats don't offer resolution information
        return "/".join(f"{f}{width}{height}{fps}/{f}{width}{height}/{f}" for f in fmt)


@dataclass(frozen=True)
class CoverOptions:
    """Settings regarding the cover image to be downloaded."""

    max_size: settings.CoverMaxSize | None


@dataclass(frozen=True)
class BackgroundOptions:
    """Settings regarding the background image to be downloaded."""

    even_with_video: bool

    def download_background(self, has_video: bool) -> bool:
        return not has_video or self.even_with_video


@dataclass(frozen=True)
class Options:
    """Settings for downloading songs."""

    song_dir: Path
    path_template: path_template.PathTemplate
    notify_discord: bool
    txt_options: TxtOptions | None
    audio_options: AudioOptions | None
    browser: settings.Browser
    video_options: VideoOptions | None
    cover: CoverOptions | None
    background_options: BackgroundOptions | None


def download_options() -> Options:
    return Options(
        song_dir=settings.get_song_dir(),
        path_template=settings.get_path_template(),
        notify_discord=settings.get_discord_allowed(),
        txt_options=_txt_options(),
        audio_options=_audio_options(),
        browser=settings.get_browser(),
        video_options=_video_options(),
        cover=_cover_options(),
        background_options=_background_options(),
    )


def _txt_options() -> TxtOptions | None:
    if not settings.get_txt():
        return None
    return TxtOptions(
        encoding=settings.get_encoding(),
        newline=settings.get_newline(),
        format_version=settings.get_version(),
        fix_linebreaks=settings.get_fix_linebreaks(),
        fix_first_words_capitalization=settings.get_fix_first_words_capitalization(),
        fix_spaces=settings.get_fix_spaces(),
        fix_quotation_marks=settings.get_fix_quotation_marks(),
    )


def _audio_options() -> AudioOptions | None:
    if not settings.get_audio():
        return None
    return AudioOptions(
        format=settings.get_audio_format(),
        bitrate=settings.get_audio_bitrate(),
        normalization=settings.get_audio_normalization(),
        embed_artwork=settings.get_audio_embed_artwork(),
        rate_limit=settings.get_ytdlp_rate_limit(),
    )


def _video_options() -> VideoOptions | None:
    if not settings.get_video():
        return None
    return VideoOptions(
        format=settings.get_video_format(),
        reencode_format=(
            settings.get_video_format_new() if settings.get_video_reencode() else None
        ),
        max_resolution=settings.get_video_resolution(),
        max_fps=settings.get_video_fps(),
        embed_artwork=settings.get_video_embed_artwork(),
        rate_limit=settings.get_ytdlp_rate_limit(),
    )


def _cover_options() -> CoverOptions | None:
    if not settings.get_cover():
        return None
    return CoverOptions(max_size=settings.get_cover_max_size() or None)


def _background_options() -> BackgroundOptions | None:
    if not settings.get_background():
        return None
    return BackgroundOptions(even_with_video=settings.get_background_always())



"""Data structure for storing user defined custom data."""

import builtins
import collections.abc
from collections import defaultdict

from usdb_syncer import db


class CustomData:
    """Dict of custom data."""

    _data: dict[str, str]
    _options: defaultdict[str, builtins.set[str]] | None = None
    FORBIDDEN_CHARACTERS = '?"<>|*.:/\\'

    @classmethod
    def value_options(cls, key: str) -> tuple[str, ...]:
        if cls._options is None:
            cls._options = db.get_custom_data_map()
        return tuple(cls._options[key])

    @classmethod
    def key_options(cls) -> tuple[str, ...]:
        if cls._options is None:
            cls._options = db.get_custom_data_map()
        return tuple(cls._options)

    @classmethod
    def is_valid_key(cls, key: str) -> bool:
        return (
            bool(key)
            and key.strip() == key
            and not any(c in key for c in cls.FORBIDDEN_CHARACTERS)
        )

    def __init__(self, data: dict[str, str] | None = None) -> None:
        self._data = data.copy() if data else {}

    def get(self, key: str) -> str | None:
        return self._data.get(key)

    def set(self, key: str, value: str | None) -> None:
        if value is None:
            if key in self._data:
                del self._data[key]
        else:
            self._data[key] = value
            if self._options is not None:
                self._options[key].add(value)

    def items(self) -> collections.abc.ItemsView[str, str]:
        return self._data.items()

    def inner(self) -> dict[str, str]:
        return self._data.copy()

    def __eq__(self, value: object) -> bool:
        return isinstance(value, CustomData) and self._data == value._data



"""Headers of a parsed UltraStar txt file."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import attrs

from usdb_syncer import errors
from usdb_syncer.logger import Logger
from usdb_syncer.meta_tags import MetaTags
from usdb_syncer.settings import FormatVersion
from usdb_syncer.song_txt.auxiliaries import BeatsPerMinute, replace_false_apostrophes
from usdb_syncer.song_txt.language_translations import LANGUAGE_TRANSLATIONS


@attrs.define
class Headers:
    """Ultrastar headers."""

    unknown: dict[str, str]
    title: str
    artist: str
    bpm: BeatsPerMinute
    gap: int = 0
    version: str | None = None
    language: str | None = None
    edition: str | None = None
    genre: str | None = None
    album: str | None = None
    year: str | None = None
    creator: str | None = None
    mp3: str | None = None
    audio: str | None = None
    audiourl: str | None = None
    vocals: str | None = None
    instrumental: str | None = None
    cover: str | None = None
    coverurl: str | None = None
    background: str | None = None
    backgroundurl: str | None = None
    video: str | None = None
    videourl: str | None = None
    videogap: float | None = None
    start: float | None = None
    end: int | None = None
    previewstart: float | None = None
    relative: str | None = None
    p1: str | None = None
    p2: str | None = None
    medleystartbeat: int | None = None
    medleyendbeat: int | None = None
    # not rewritten, as it depends on the chosen encoding
    encoding: str | None = None
    comment: str | None = None
    resolution: str | None = None
    tags: str | None = None

    @classmethod
    def parse(cls, lines: list[str], logger: Logger) -> Headers:
        """Consumes a stream of lines while they are headers."""
        kwargs: dict[str, Any] = {"unknown": {}}
        while lines:
            if not lines[0].startswith("#"):
                break
            line = lines.pop(0).removeprefix("#")
            if ":" not in line:
                logger.warning(f"header without value: '{line}'")
                continue
            header, value = line.split(":", maxsplit=1)
            if not value:
                # ignore headers with empty values
                continue
            try:
                _set_header_value(kwargs, header, value)
            except ValueError:
                logger.warning(f"invalid header value: '{line}'")
        if "title" not in kwargs or "artist" not in kwargs or "bpm" not in kwargs:
            raise errors.HeadersRequiredMissingError()
        return cls(**kwargs)

    def set_version(self, version: FormatVersion) -> None:
        self.version = version.value

    def reset_file_location_headers(self) -> None:
        """Clear all tags with local file locations."""
        self.mp3 = self.audio = self.vocals = self.instrumental = self.video = (
            self.cover
        ) = self.background = None

    def __str__(self) -> str:
        out = "\n".join(
            f"#{key.upper()}:{val}"
            for key in (
                "version",
                "title",
                "artist",
                "language",
                "edition",
                "genre",
                "year",
                "creator",
                "mp3",
                "audio",
                "audiourl",
                "vocals",
                "instrumental",
                "cover",
                "coverurl",
                "background",
                "backgroundurl",
                "video",
                "videourl",
                "videogap",
                "resolution",
                "start",
                "end",
                "relative",
                "previewstart",
                "medleystartbeat",
                "medleyendbeat",
                "bpm",
                "gap",
                "p1",
                "p2",
                "album",
                "comment",
                "tags",
            )
            if (val := getattr(self, key)) is not None
        )
        if self.unknown:
            out = "\n".join((
                out,
                *(f"#{key.upper()}:{val}" for key, val in self.unknown.items()),
            ))
        return out

    def artist_title_str(self) -> str:
        return f"{self.artist} - {self.title}"

    def fix_apostrophes(self, logger: Logger) -> None:
        apostrophes_and_quotation_marks_fixed = False
        for key in ("artist", "title", "language", "genre", "p1", "p2", "album"):
            if value := getattr(self, key):
                corrected_value = replace_false_apostrophes(value)
                if value != corrected_value:
                    setattr(self, key, corrected_value)
                    apostrophes_and_quotation_marks_fixed = True
        if apostrophes_and_quotation_marks_fixed:
            logger.debug("FIX: Apostrophes in song header corrected.")

    def apply_to_medley_tags(self, func: Callable[[int], int]) -> None:
        if self.medleystartbeat:
            self.medleystartbeat = func(self.medleystartbeat)
        if self.medleyendbeat:
            self.medleyendbeat = func(self.medleyendbeat)

    def fix_language(self, logger: Logger) -> None:
        if not self.language:
            logger.debug("No #LANGUAGE tag found. Consider adding it.")
        if old_language := self.language:
            languages = [
                language.strip()
                for language in self.language.replace(";", ",")
                .replace("/", ",")
                .replace("|", ",")
                .split(",")
            ]
            languages = [
                LANGUAGE_TRANSLATIONS.get(language.lower(), language)
                for language in languages
            ]
            self.language = ", ".join(languages)
            if old_language != self.language:
                logger.debug(f"FIX: Language corrected to {self.language}.")

    def main_language(self) -> str:
        if self.language:
            return self.language.split(",", maxsplit=1)[0].removesuffix(" (romanized)")
        return ""

    def fix_videogap(self, meta_tags: MetaTags, logger: Logger) -> None:
        if self.videogap is not None:
            if (
                meta_tags.audio is None
                or meta_tags.video is None
                or meta_tags.audio == meta_tags.video
            ):
                logger.warning(
                    "This song contains a non-zero #VIDEOGAP, which only makes sense "
                    "if different resources for audio and video are specified, which "
                    "is not the case here. This should be fixed in USDB. Removing "
                    "#VIDEOGAP in local text file."
                )
                self.videogap = None


def _set_header_value(kwargs: dict[str, Any], header: str, value: str) -> None:
    header = "creator" if header == "AUTHOR" else header.lower()
    if header in (
        "artist",
        "version",
        "language",
        "edition",
        "genre",
        "album",
        "year",
        "creator",
        "mp3",
        "audio",
        "audiourl",
        "vocals",
        "instrumental",
        "cover",
        "coverurl",
        "background",
        "backgroundurl",
        "video",
        "videourl",
        "relative",
        "p1",
        "p2",
        "encoding",
        "comment",
        "resolution",
        "tags",
    ):
        kwargs[header] = value
    elif header == "title":
        kwargs[header] = value.removesuffix(" [DUET]")
    # these are given in (fractional) seconds, thus may have a decimal comma or point
    elif header in ("videogap", "start", "previewstart"):
        if (val := float(value.replace(",", "."))) != 0.0:
            kwargs[header] = val
    # these are given in milliseconds, but may have a decimal point or comma in usdb
    elif header in ("gap", "end"):
        kwargs[header] = round(float(value.replace(",", ".")))
    # these are given in beats and should thus be integers
    elif header in ("medleystartbeat", "medleyendbeat"):
        kwargs[header] = int(value)
    elif header == "bpm":
        kwargs[header] = BeatsPerMinute.parse(value)
    else:
        kwargs["unknown"][header] = value



"""Controller for the song table."""

from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator
from functools import partial
from typing import TYPE_CHECKING, Any

import send2trash
from PySide6 import QtCore, QtMultimedia, QtWidgets
from PySide6.QtCore import QItemSelectionModel, Qt
from PySide6.QtGui import QAction, QCursor, QKeySequence, QShortcut

from usdb_syncer import SongId, db, events, media_player, settings, sync_meta
from usdb_syncer.gui import ffmpeg_dialog
from usdb_syncer.gui.custom_data_dialog import CustomDataDialog
from usdb_syncer.gui.progress import run_with_progress
from usdb_syncer.gui.song_table.column import Column
from usdb_syncer.gui.song_table.table_model import TableModel
from usdb_syncer.logger import song_logger
from usdb_syncer.song_loader import DownloadManager
from usdb_syncer.usdb_song import DownloadStatus, UsdbSong

if TYPE_CHECKING:
    from usdb_syncer.gui.mw import MainWindow


DEFAULT_COLUMN_WIDTH = 300


class SongTable:
    """Controller for the song table."""

    _search = db.SearchBuilder()
    _playing_song: UsdbSong | None = None
    _next_playing_song: UsdbSong | None = None

    def __init__(self, mw: MainWindow) -> None:
        self.mw = mw
        self._model = TableModel(mw)
        self._view = mw.table_view
        self._media_player = media_player.media_player()
        self._media_player.playbackStateChanged.connect(self._on_playback_state_changed)
        self._media_player.errorChanged.connect(self._on_playback_error_changed)
        self._setup_view()
        self._header().sortIndicatorChanged.connect(self._on_sort_order_changed)
        mw.table_view.selectionModel().currentChanged.connect(
            self._on_current_song_changed
        )
        events.SongChanged.subscribe(self._on_song_changed)
        self._setup_search_timer()
        events.TreeFilterChanged.subscribe(self._on_tree_filter_changed)
        events.TextFilterChanged.subscribe(self._on_text_filter_changed)
        events.SavedSearchRestored.subscribe(self._on_saved_search_restored)
        QShortcut(QKeySequence(Qt.Key.Key_Space), self._view).activated.connect(
            self._on_space
        )

    def _on_playback_state_changed(
        self, state: QtMultimedia.QMediaPlayer.PlaybackState
    ) -> None:
        if state == QtMultimedia.QMediaPlayer.PlaybackState.PlayingState:
            assert self._next_playing_song
            self._playing_song = song = self._next_playing_song
            self._next_playing_song = None
            song.is_playing = True
        else:
            assert self._playing_song
            song = self._playing_song
            self._playing_song = None
            song.is_playing = False
        with db.transaction():
            song.upsert()
        events.SongChanged(song.song_id).post()

    def _on_playback_error_changed(self) -> None:
        if not self._media_player.error().value or self._next_playing_song is None:
            return
        logger = song_logger(self._next_playing_song.song_id)
        source = self._media_player.source().url()
        logger.error(f"Failed to play back source: {source}")
        logger.debug(self._media_player.errorString())

    def _on_space(self) -> None:
        if song := self.current_song():
            self._play_or_stop_sample(song)

    def _header(self) -> QtWidgets.QHeaderView:
        return self._view.horizontalHeader()

    def reset(self) -> None:
        self._model.reset()

    def begin_reset(self) -> None:
        self._model.beginResetModel()

    def end_reset(self) -> None:
        self._model.endResetModel()

    def download_selection(self) -> None:
        self._download(self._selected_rows())

    def _download(self, rows: Iterable[int]) -> None:
        ffmpeg_dialog.check_ffmpeg(
            self.mw,
            lambda: run_with_progress(
                "Initializing downloads ...", partial(self._download_inner, rows)
            ),
        )

    def _download_inner(self, rows: Iterable[int]) -> None:
        to_download: list[UsdbSong] = []
        for song_id in self._model.ids_for_rows(rows):
            song = UsdbSong.get(song_id)
            assert song
            if song.sync_meta and song.sync_meta.pinned:
                song_logger(song.song_id).info("Not downloading song as it is pinned.")
                continue
            if song.status.can_be_downloaded():
                song.status = DownloadStatus.PENDING
                with db.transaction():
                    song.upsert()
                events.SongChanged(song.song_id).post()
                to_download.append(song)
        if to_download:
            events.DownloadsRequested(len(to_download)).post()
            DownloadManager.download(to_download)

    def abort_selected_downloads(self) -> None:
        ids = self._model.ids_for_rows(self._selected_rows())
        run_with_progress("Aborting downloads ...", lambda: DownloadManager.abort(ids))

    def _setup_view(self) -> None:
        state = settings.get_table_view_header_state()
        self._view.setModel(self._model)
        self._view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self._view.customContextMenuRequested.connect(
            lambda: self.mw.menu_songs.exec(QCursor.pos())
        )
        self._view.doubleClicked.connect(lambda idx: self._download([idx.row()]))
        self._view.clicked.connect(self._on_click)
        header = self._header()
        existing_state = False
        if not state.isEmpty():
            header.restoreState(state)
            if header.count() != max(Column) + 1:
                header.reset()
            else:
                existing_state = True
                self._search.order = Column(header.sortIndicatorSection()).song_order()
                self._search.descending = bool(header.sortIndicatorOrder().value)
        for column in Column:
            if size := column.fixed_size():
                header.setSectionResizeMode(
                    column, QtWidgets.QHeaderView.ResizeMode.Fixed
                )
                header.resizeSection(column, size)
            # setting a (default) width on the last, stretching column seems to cause
            # issues, so we set it manually on the other columns
            elif column is not max(Column):
                if not existing_state:
                    header.resizeSection(column, DEFAULT_COLUMN_WIDTH)
                header.setSectionResizeMode(
                    column, QtWidgets.QHeaderView.ResizeMode.Interactive
                )

    def build_custom_data_menu(self) -> None:
        if not (song := self.current_song()) or not song.sync_meta:
            return

        def update(key: str, value: str | None) -> None:
            if not song.sync_meta:
                return
            song.sync_meta.custom_data.set(key, value)
            with db.transaction():
                song.upsert()
            events.SongChanged(song.song_id).post()

        def run_custom_data_dialog(key: str | None = None) -> None:
            CustomDataDialog(self.mw, update, key).open()

        self.mw.menu_custom_data.clear()
        _add_action("New ...", self.mw.menu_custom_data, run_custom_data_dialog)
        self.mw.menu_custom_data.addSeparator()
        for key, value in song.sync_meta.custom_data.items():
            key_menu = QtWidgets.QMenu(key, self.mw.menu_custom_data)
            self.mw.menu_custom_data.addMenu(key_menu)
            _add_action("New ...", key_menu, partial(run_custom_data_dialog, key))
            key_menu.addSeparator()
            _add_action(value, key_menu, partial(update, key, None), checked=True)
            for option in sync_meta.CustomData.value_options(key):
                if option != value:
                    _add_action(
                        option, key_menu, partial(update, key, option), checked=False
                    )

    def _on_click(self, index: QtCore.QModelIndex) -> None:
        if index.column() == Column.SAMPLE_URL.value and (
            song := UsdbSong.get(self._model.ids_for_indices([index])[0])
        ):
            self._play_or_stop_sample(song)

    def _play_or_stop_sample(self, song: UsdbSong) -> None:
        if self._playing_song and song.song_id == self._playing_song.song_id:
            # second play() in a row is a stop()
            self._media_player.stop()
            return
        position = 0
        if song.sync_meta and song.sync_meta.audio:
            path = song.sync_meta.path.parent / song.sync_meta.audio.fname
            url = path.absolute().as_posix()
            if song.sync_meta.meta_tags.preview:
                position = int(song.sync_meta.meta_tags.preview * 1000)
        elif song.sample_url:
            url = song.sample_url
        else:
            return
        self._next_playing_song = song
        self._media_player.setSource(url)
        self._media_player.setPosition(position)
        self._media_player.play()

    def save_state(self) -> None:
        settings.set_table_view_header_state(
            self.mw.table_view.horizontalHeader().saveState()
        )

    # actions

    def _on_song_changed(self, event: events.SongChanged) -> None:
        if event.song_id == self.current_song_id():
            self._on_current_song_changed()

    def _on_current_song_changed(self) -> None:
        song = self.current_song()
        for action in self.mw.menu_songs.actions():
            action.setEnabled(bool(song))
        if not song:
            return
        for action in (
            self.mw.action_open_song_folder,
            self.mw.menu_open_song_in,
            self.mw.action_open_song_in_karedi,
            self.mw.action_open_song_in_performous,
            self.mw.action_open_song_in_ultrastar_manager,
            self.mw.action_open_song_in_usdx,
            self.mw.action_open_song_in_vocaluxe,
            self.mw.action_open_song_in_yass_reloaded,
            self.mw.action_delete,
            self.mw.action_pin,
            self.mw.menu_custom_data,
        ):
            action.setEnabled(song.is_local())
        self.mw.action_open_song_in_karedi.setVisible(
            settings.get_app_path(settings.SupportedApps.KAREDI) is not None
        )
        self.mw.action_open_song_in_performous.setVisible(
            settings.get_app_path(settings.SupportedApps.PERFORMOUS) is not None
        )
        self.mw.action_open_song_in_ultrastar_manager.setVisible(
            settings.get_app_path(settings.SupportedApps.ULTRASTAR_MANAGER) is not None
        )
        self.mw.action_open_song_in_usdx.setVisible(
            settings.get_app_path(settings.SupportedApps.USDX) is not None
        )
        self.mw.action_open_song_in_vocaluxe.setVisible(
            settings.get_app_path(settings.SupportedApps.VOCALUXE) is not None
        )
        self.mw.action_open_song_in_yass_reloaded.setVisible(
            settings.get_app_path(settings.SupportedApps.YASS_RELOADED) is not None
        )
        self.mw.action_pin.setChecked(song.is_pinned())
        self.mw.action_songs_abort.setEnabled(song.status.can_be_aborted())

    def delete_selected_songs(self) -> None:
        with db.transaction():
            for song in self.selected_songs():
                if not song.sync_meta:
                    continue
                logger = song_logger(song.song_id)
                if song.is_pinned():
                    logger.info("Not trashing song folder as it is pinned.")
                    continue
                if song.sync_meta.path.exists():
                    send2trash.send2trash(song.sync_meta.path.parent)
                song.remove_sync_meta()
                events.SongChanged(song.song_id)
                logger.debug("Trashed song folder.")

    def set_pin_selected_songs(self, pin: bool) -> None:
        for song in self.selected_songs():
            if not song.sync_meta or song.sync_meta.pinned == pin:
                continue
            song.sync_meta.pinned = pin
            song.sync_meta.synchronize_to_file()
            with db.transaction():
                song.sync_meta.upsert()
            events.SongChanged(song.song_id)

    # selection model

    def current_song_id(self) -> SongId | None:
        if (idx := self._view.selectionModel().currentIndex()).isValid():
            if ids := self._model.ids_for_indices([idx]):
                return ids[0]
        return None

    def current_song(self) -> UsdbSong | None:
        if (song_id := self.current_song_id()) is not None:
            return UsdbSong.get(song_id)
        return None

    def selected_songs(self) -> Iterator[UsdbSong]:
        return (
            song
            for song_id in self._model.ids_for_rows(self._selected_rows())
            if (song := UsdbSong.get(song_id))
        )

    def _selected_rows(self) -> Iterable[int]:
        return (idx.row() for idx in self._view.selectionModel().selectedRows())

    def set_selection_to_song_ids(self, select_song_ids: Iterable[SongId]) -> None:
        self.set_selection_to_indices(self._model.indices_for_ids(select_song_ids))

    def set_selection_to_indices(self, rows: Iterable[QtCore.QModelIndex]) -> None:
        selection = QtCore.QItemSelection()
        for row in rows:
            selection.select(row, row)
        self._view.selectionModel().select(
            selection,
            QItemSelectionModel.SelectionFlag.Rows
            | QItemSelectionModel.SelectionFlag.ClearAndSelect,
        )

    # sorting and filtering

    def connect_row_count_changed(self, func: Callable[[int, int], None]) -> None:
        """Calls `func` with the new table row and selection counts."""

        def wrapped(*_: Any) -> None:
            func(
                self._model.rowCount(), len(self._view.selectionModel().selectedRows())
            )

        self._model.modelReset.connect(wrapped)
        self._model.rowsInserted.connect(wrapped)
        self._model.rowsRemoved.connect(wrapped)
        self._view.selectionModel().selectionChanged.connect(wrapped)

    def _setup_search_timer(self) -> None:
        self._search_timer = QtCore.QTimer(self.mw)
        self._search_timer.setSingleShot(True)
        self._search_timer.timeout.connect(self.search_songs)

    def search_songs(self, msec_delay: int = 0) -> None:
        self._search_timer.stop()
        if msec_delay:
            self._search_timer.setInterval(msec_delay)
            self._search_timer.start()
        else:
            self._model.set_songs(db.search_usdb_songs(self._search))
            self._on_current_song_changed()

    def _on_tree_filter_changed(self, event: events.TreeFilterChanged) -> None:
        event.search.order = self._search.order
        event.search.descending = self._search.descending
        event.search.text = self._search.text
        self._search = event.search
        self.search_songs(100)

    def _on_saved_search_restored(self, event: events.SavedSearchRestored) -> None:
        self._search.order = event.search.order
        self._search.descending = event.search.descending
        self._search.text = event.search.text
        self._header().setSortIndicator(
            Column.from_song_order(event.search.order),
            (
                Qt.SortOrder.DescendingOrder
                if event.search.descending
                else Qt.SortOrder.AscendingOrder
            ),
        )
        self.search_songs(100)

    def _on_text_filter_changed(self, event: events.TextFilterChanged) -> None:
        self._search.text = event.search
        self.search_songs(400)

    def _on_sort_order_changed(self, section: int, order: Qt.SortOrder) -> None:
        self._search.order = Column(section).song_order()
        self._search.descending = bool(order.value)
        events.SearchOrderChanged(self._search.order, self._search.descending).post()
        self.search_songs()


def _add_action(
    name: str,
    menu: QtWidgets.QMenu,
    slot: Callable[[], None],
    checked: bool | None = None,
) -> None:
    action = QAction(name, menu)
    if checked is not None:
        action.setCheckable(True)
        action.setChecked(checked)
    action.triggered.connect(slot)
    menu.addAction(action)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'FfmpegDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QHBoxLayout, QLabel, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.setWindowModality(Qt.WindowModality.ApplicationModal)
        Dialog.resize(624, 194)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label_ffmpeg = QLabel(Dialog)
        self.label_ffmpeg.setObjectName(u"label_ffmpeg")
        self.label_ffmpeg.setTextFormat(Qt.TextFormat.RichText)
        self.label_ffmpeg.setWordWrap(True)

        self.verticalLayout.addWidget(self.label_ffmpeg)

        self.label_windows = QLabel(Dialog)
        self.label_windows.setObjectName(u"label_windows")
        self.label_windows.setEnabled(True)
        self.label_windows.setTextFormat(Qt.TextFormat.RichText)
        self.label_windows.setWordWrap(True)
        self.label_windows.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextSelectableByMouse)

        self.verticalLayout.addWidget(self.label_windows)

        self.label_macos = QLabel(Dialog)
        self.label_macos.setObjectName(u"label_macos")
        self.label_macos.setEnabled(True)
        self.label_macos.setTextFormat(Qt.TextFormat.RichText)
        self.label_macos.setWordWrap(True)
        self.label_macos.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextSelectableByMouse)

        self.verticalLayout.addWidget(self.label_macos)

        self.label_linux = QLabel(Dialog)
        self.label_linux.setObjectName(u"label_linux")
        self.label_linux.setEnabled(True)
        self.label_linux.setTextFormat(Qt.TextFormat.RichText)
        self.label_linux.setWordWrap(True)

        self.verticalLayout.addWidget(self.label_linux)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.set_location = QPushButton(Dialog)
        self.set_location.setObjectName(u"set_location")

        self.horizontalLayout.addWidget(self.set_location)

        self.buttonBox = QDialogButtonBox(Dialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Close)

        self.horizontalLayout.addWidget(self.buttonBox)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"FFmpeg not found", None))
        self.label_ffmpeg.setText(QCoreApplication.translate("Dialog", u"<html><head/><body><p><span style=\" font-weight:700;\">FFmpeg</span> was not found, but is required to download resources!</p></body></html>", None))
        self.label_windows.setText(QCoreApplication.translate("Dialog", u"<html><head/><body><p>The recommended way to install FFmpeg is to use a package manager such as <a href=\"https://chocolatey.org/\"><span style=\" text-decoration: underline; color:#2e54ff;\">Chocolatey</span></a> (<span style=\" font-family:'Courier New';\">choco install ffmpeg</span>) or <a href=\"https://learn.microsoft.com/en-us/windows/package-manager/winget/\"><span style=\" text-decoration: underline; color:#2e54ff;\">winget</span></a> (<span style=\" font-family:'Courier New';\">winget install --id=Gyan.FFmpeg -e</span>).</p><p>Alternatively, download the FFmpeg binaries from <a href=\"https://www.gyan.dev/ffmpeg/builds/ffmpeg-git-full.7z\"><span style=\" text-decoration: underline; color:#2e54ff;\">https://www.gyan.dev/ffmpeg/builds/ffmpeg-git-full.7z</span></a>, extract the archive and then set the location with the button below. </p></body></html>", None))
        self.label_macos.setText(QCoreApplication.translate("Dialog", u"<html><head/><body><p>The recommended way to install FFmpeg is via <a href=\"https://brew.sh/\"><span style=\" text-decoration: underline; color:#2e54ff;\">Homebrew</span></a> (<span style=\" font-family:'Courier New';\">brew install ffmpeg</span>).</p></body></html>", None))
        self.label_linux.setText(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Please install the FFmpeg package through your distribution\u2019s package manager.</p></body></html>", None))
        self.set_location.setText(QCoreApplication.translate("Dialog", u"Set Location...", None))
    # retranslateUi




"""Table model for song data."""

from __future__ import annotations

import enum
from enum import IntEnum
from typing import assert_never

from PySide6.QtGui import QIcon

from usdb_syncer import db
from usdb_syncer.gui.icons import Icon


class Column(IntEnum):
    """Table columns."""

    SAMPLE_URL = 0
    SONG_ID = enum.auto()
    ARTIST = enum.auto()
    TITLE = enum.auto()
    LANGUAGE = enum.auto()
    EDITION = enum.auto()
    GOLDEN_NOTES = enum.auto()
    RATING = enum.auto()
    VIEWS = enum.auto()
    YEAR = enum.auto()
    GENRE = enum.auto()
    CREATOR = enum.auto()
    TAGS = enum.auto()
    PINNED = enum.auto()
    TXT = enum.auto()
    AUDIO = enum.auto()
    VIDEO = enum.auto()
    COVER = enum.auto()
    BACKGROUND = enum.auto()
    DOWNLOAD_STATUS = enum.auto()

    def display_data(self) -> str | None:  # noqa: C901
        match self:
            case Column.SONG_ID:
                return "ID"
            case Column.ARTIST:
                return "Artist"
            case Column.TITLE:
                return "Title"
            case Column.LANGUAGE:
                return "Language"
            case Column.EDITION:
                return "Edition"
            case Column.YEAR:
                return "Year"
            case Column.GENRE:
                return "Genre"
            case Column.CREATOR:
                return "Creator"
            case Column.TAGS:
                return "Tags"
            case Column.DOWNLOAD_STATUS:
                return "Status"
            case (
                Column.SAMPLE_URL
                | Column.GOLDEN_NOTES
                | Column.RATING
                | Column.VIEWS
                | Column.TXT
                | Column.AUDIO
                | Column.VIDEO
                | Column.COVER
                | Column.BACKGROUND
                | Column.PINNED
            ):
                return None
            case _ as unreachable:
                assert_never(unreachable)

    def decoration_data(self) -> QIcon:  # noqa: C901
        match self:
            case Column.SAMPLE_URL:
                icon = Icon.AUDIO_SAMPLE
            case Column.SONG_ID:
                icon = Icon.ID
            case Column.ARTIST:
                icon = Icon.ARTIST
            case Column.TITLE:
                icon = Icon.TITLE
            case Column.LANGUAGE:
                icon = Icon.LANGUAGE
            case Column.EDITION:
                icon = Icon.EDITION
            case Column.GOLDEN_NOTES:
                icon = Icon.GOLDEN_NOTES
            case Column.RATING:
                icon = Icon.RATING
            case Column.VIEWS:
                icon = Icon.VIEWS
            case Column.YEAR:
                icon = Icon.CALENDAR
            case Column.GENRE:
                icon = Icon.GENRE
            case Column.CREATOR:
                icon = Icon.CREATOR
            case Column.TAGS:
                icon = Icon.TAGS
            case Column.TXT:
                icon = Icon.TEXT
            case Column.AUDIO:
                icon = Icon.AUDIO
            case Column.VIDEO:
                icon = Icon.VIDEO
            case Column.COVER:
                icon = Icon.COVER
            case Column.BACKGROUND:
                icon = Icon.BACKGROUND
            case Column.DOWNLOAD_STATUS:
                icon = Icon.DOWNLOAD
            case Column.PINNED:
                icon = Icon.PIN
            case _ as unreachable:
                assert_never(unreachable)
        return icon.icon()

    def fixed_size(self) -> int | None:
        match self:
            case (
                Column.ARTIST
                | Column.TITLE
                | Column.LANGUAGE
                | Column.EDITION
                | Column.DOWNLOAD_STATUS
                | Column.SONG_ID
                | Column.VIEWS
                | Column.RATING
                | Column.GOLDEN_NOTES
                | Column.YEAR
                | Column.GENRE
                | Column.CREATOR
                | Column.TAGS
            ):
                return None
            case (
                Column.SAMPLE_URL
                | Column.TXT
                | Column.AUDIO
                | Column.VIDEO
                | Column.COVER
                | Column.BACKGROUND
                | Column.PINNED
            ):
                return 24
            case _ as unreachable:
                assert_never(unreachable)

    def song_order(self) -> db.SongOrder:  # noqa: C901
        match self:
            case Column.SAMPLE_URL:
                return db.SongOrder.SAMPLE_URL
            case Column.SONG_ID:
                return db.SongOrder.SONG_ID
            case Column.ARTIST:
                return db.SongOrder.ARTIST
            case Column.TITLE:
                return db.SongOrder.TITLE
            case Column.LANGUAGE:
                return db.SongOrder.LANGUAGE
            case Column.EDITION:
                return db.SongOrder.EDITION
            case Column.GOLDEN_NOTES:
                return db.SongOrder.GOLDEN_NOTES
            case Column.RATING:
                return db.SongOrder.RATING
            case Column.VIEWS:
                return db.SongOrder.VIEWS
            case Column.YEAR:
                return db.SongOrder.YEAR
            case Column.GENRE:
                return db.SongOrder.GENRE
            case Column.CREATOR:
                return db.SongOrder.CREATOR
            case Column.TAGS:
                return db.SongOrder.TAGS
            case Column.PINNED:
                return db.SongOrder.PINNED
            case Column.TXT:
                return db.SongOrder.TXT
            case Column.AUDIO:
                return db.SongOrder.AUDIO
            case Column.VIDEO:
                return db.SongOrder.VIDEO
            case Column.COVER:
                return db.SongOrder.COVER
            case Column.BACKGROUND:
                return db.SongOrder.BACKGROUND
            case Column.DOWNLOAD_STATUS:
                return db.SongOrder.STATUS
            case unreachable:
                assert_never(unreachable)

    @classmethod
    def from_song_order(cls, song_order: db.SongOrder) -> Column:  # noqa: C901
        match song_order:
            case db.SongOrder.SAMPLE_URL:
                return Column.SAMPLE_URL
            case db.SongOrder.SONG_ID | db.SongOrder.NONE:
                return Column.SONG_ID
            case db.SongOrder.ARTIST:
                return Column.ARTIST
            case db.SongOrder.TITLE:
                return Column.TITLE
            case db.SongOrder.LANGUAGE:
                return Column.LANGUAGE
            case db.SongOrder.EDITION:
                return Column.EDITION
            case db.SongOrder.GOLDEN_NOTES:
                return Column.GOLDEN_NOTES
            case db.SongOrder.RATING:
                return Column.RATING
            case db.SongOrder.VIEWS:
                return Column.VIEWS
            case db.SongOrder.YEAR:
                return Column.YEAR
            case db.SongOrder.GENRE:
                return Column.GENRE
            case db.SongOrder.CREATOR:
                return Column.CREATOR
            case db.SongOrder.TAGS:
                return Column.TAGS
            case db.SongOrder.PINNED:
                return Column.PINNED
            case db.SongOrder.TXT:
                return Column.TXT
            case db.SongOrder.AUDIO:
                return Column.AUDIO
            case db.SongOrder.VIDEO:
                return Column.VIDEO
            case db.SongOrder.COVER:
                return Column.COVER
            case db.SongOrder.BACKGROUND:
                return Column.BACKGROUND
            case db.SongOrder.STATUS:
                return Column.DOWNLOAD_STATUS
            case unreachable:
                assert_never(unreachable)



"""Hooks called from the GUI that add-ons can subscribe to."""

from __future__ import annotations

from typing import TYPE_CHECKING

from usdb_syncer import hooks

if TYPE_CHECKING:
    from usdb_syncer.gui.mw import MainWindow


class MainWindowDidLoad(hooks._Hook):
    """Called after the main window has loaded."""

    @classmethod
    def call(cls, main_window: MainWindow) -> None:
        super().call(main_window)



"""High-level routines for USDB and local songs."""

import json
import os
from collections.abc import Generator
from importlib import resources
from importlib.abc import Traversable
from pathlib import Path

import requests
import send2trash
from requests import Session

from usdb_syncer import (
    SongId,
    SyncMetaId,
    data,
    db,
    errors,
    events,
    settings,
    song_txt,
    utils,
)
from usdb_syncer.logger import error_logger, logger
from usdb_syncer.song_loader import DownloadManager
from usdb_syncer.sync_meta import SyncMeta
from usdb_syncer.usdb_scraper import get_usdb_available_songs
from usdb_syncer.usdb_song import UsdbSong, UsdbSongEncoder
from usdb_syncer.utils import AppPaths


def load_available_songs(force_reload: bool, session: Session | None = None) -> None:
    if force_reload:
        max_skip_id = SongId(0)
        UsdbSong.delete_all()
    elif (max_skip_id := db.max_usdb_song_id()) == 0 and (songs := load_cached_songs()):
        UsdbSong.upsert_many(songs)
        max_skip_id = db.max_usdb_song_id()
    try:
        songs = get_usdb_available_songs(max_skip_id, session=session)
    except errors.UsdbLoginError:
        logger.debug("Skipping fetching new songs as there is no login.")
        return
    except requests.exceptions.ConnectionError:
        logger.debug("", exc_info=True)
        logger.error("Failed to fetch new songs; check network connection.")
        return
    if songs:
        UsdbSong.upsert_many(songs)
        _download_subscribed_songs(songs)


def _download_subscribed_songs(songs: list[UsdbSong]) -> None:
    if not settings.ffmpeg_is_available():
        return
    subscribed = set(db.SavedSearch.get_subscribed_song_ids())
    to_download = []
    if subscribed:
        for song in songs:
            if song.song_id in subscribed:
                song.status = db.DownloadStatus.PENDING
                UsdbSong.upsert(song)
                to_download.append(song)
    if to_download:
        events.DownloadsRequested(len(to_download)).post()
        DownloadManager.download(to_download)


def load_cached_songs() -> list[UsdbSong] | None:
    if AppPaths.song_list.exists():
        path: Path | Traversable = AppPaths.song_list
    elif (resource := resources.files(data).joinpath("song_list.json")).is_file():
        path = resource
    else:
        return None
    with path.open(encoding="utf8") as file:
        try:
            return json.load(file, object_hook=UsdbSong.from_json)
        except (json.decoder.JSONDecodeError, TypeError, KeyError):
            return None


def dump_available_songs(songs: list[UsdbSong], target: Path | None = None) -> None:
    with (target or AppPaths.song_list).open("w", encoding="utf8") as file:
        json.dump(songs, file, cls=UsdbSongEncoder)


def _iterate_usdb_files_in_folder_recursively(
    folder: Path,
) -> Generator[Path, None, None]:
    for root, _, files in os.walk(folder):
        for file in files:
            if file.endswith(".usdb") and not file.startswith("."):
                yield Path(root) / file


def synchronize_sync_meta_folder(folder: Path) -> None:
    db_metas = {m.sync_meta_id: m for m in SyncMeta.get_in_folder(folder)}
    song_ids = set(db.all_song_ids())
    to_upsert: list[SyncMeta] = []
    found_metas: set[SyncMetaId] = set()

    for path in _iterate_usdb_files_in_folder_recursively(folder=folder):
        if meta_id := SyncMetaId.from_path(path):
            if meta_id in found_metas:
                send2trash.send2trash(path)
                logger.warning(f"Trashed duplicated meta file: '{path}'")
                continue
            found_metas.add(meta_id)

        meta = None if meta_id is None else db_metas.get(meta_id)

        if meta_id is not None and meta and meta.mtime == utils.get_mtime(path):
            # file is unchanged
            if not utils.compare_unicode_paths(path, meta.path):
                meta.path = path
                to_upsert.append(meta)
                logger.info(f"Meta file was moved: '{path}'.")
            continue

        if meta := SyncMeta.try_from_file(path):
            if meta.song_id in song_ids:
                # file was changed and maybe moved
                to_upsert.append(meta)
                if meta.sync_meta_id in db_metas:
                    logger.info(f"Updated meta file from disk: '{path}'.")
                else:
                    logger.info(f"New meta file found on disk: '{path}'.")
            else:
                logger.info(
                    f"{meta.song_id.usdb_detail_url()} no longer exists: '{path}'."
                )

    SyncMeta.delete_many(tuple(db_metas.keys() - found_metas))
    SyncMeta.upsert_many(to_upsert)


def find_local_songs(directory: Path) -> set[SongId]:
    matched_rows: set[SongId] = set()
    for path in directory.glob("**/*.txt"):
        if headers := try_parse_txt_headers(path):
            name = headers.artist_title_str()
            if matches := list(
                db.find_similar_usdb_songs(
                    utils.normalize(headers.artist), utils.normalize(headers.title)
                )
            ):
                plural = "es" if len(matches) > 1 else ""
                logger.info(f"{len(matches)} match{plural} for '{name}'.")
                matched_rows.update(matches)
            else:
                logger.warning(f"No matches for '{name}'.")
    return matched_rows


def try_parse_txt_headers(path: Path) -> song_txt.Headers | None:
    if lines := utils.read_file_head(path, 20):
        try:
            return song_txt.Headers.parse(lines, error_logger)
        except errors.HeadersParseError:
            return None
    return None



"""Helpers for logging.

Logging levels:
    CRITICAL:   The app is forced to shut down or will stop working as intended.
    ERROR:      A problem occured comprimising the outcome of an operation.
    WARNING:    A problem occured, but the app has a way to recover.
    INFO:       Something happened that is of interest to the user.
    DEBUG:      Something happened that might be relevant for diagnosis or debugging.
"""

import logging
from types import TracebackType
from typing import Any

from usdb_syncer import SongId


class Logger(logging.LoggerAdapter):
    """Logger wrapper with our custom logic."""

    def exception(
        self,
        msg: object,
        *args: object,
        exc_info: (
            bool
            | tuple[type[BaseException], BaseException, TracebackType | None]
            | tuple[None, None, None]
            | BaseException
            | None
        ) = True,
        **kwargs: Any,
    ) -> None:
        """Log exception info with debug and message with error level."""
        if exc_info:
            self.debug(None, exc_info=exc_info, **kwargs)
        self.error(msg, *args, exc_info=False, **kwargs)


class SongLogger(Logger):
    """Logger wrapper that takes care of logging the song id."""

    def __init__(self, song_id: SongId, logger_: Any, extra: Any = ...) -> None:
        super().__init__(logger_, extra)
        self.song_id = song_id

    def process(self, msg: str, kwargs: Any) -> Any:
        return f"#{self.song_id}: {msg}", kwargs


_LOGGER_NAME = "usdb_syncer"
_raw_logger = logging.getLogger(_LOGGER_NAME)
logger = Logger(_raw_logger)
error_logger = Logger(_raw_logger.getChild("errors"))
error_logger.setLevel(logging.ERROR)


def song_logger(song_id: SongId) -> SongLogger:
    return SongLogger(song_id, logger)


def configure_logging(*handlers: logging.Handler) -> None:
    logging.basicConfig(
        level=logging.INFO,
        style="{",
        format="{asctime} [{levelname}] {message}",
        datefmt="%Y-%m-%d %H:%M:%S",
        encoding="utf-8",
        handlers=handlers,
    )
    logger.setLevel(logging.DEBUG)
